"""Sidebar for the main window"""

import locale
from gettext import gettext as _
from typing import Callable, List, Optional, Set, Tuple, Union

from gi.repository import Gdk, GObject, Gtk, Pango

from lutris import runners, services, settings
from lutris.config import LutrisConfig
from lutris.database import categories as categories_db
from lutris.database import games as games_db
from lutris.database import saved_searches as saved_search_db
from lutris.database.categories import CATEGORIES_UPDATED
from lutris.database.saved_searches import SAVED_SEARCHES_UPDATED
from lutris.game import GAME_START, GAME_STOPPED, GAME_UPDATED, Game
from lutris.gui.config.edit_category_games import EditCategoryGamesDialog
from lutris.gui.config.edit_saved_search import EditSavedSearchDialog
from lutris.gui.config.runner import RunnerConfigDialog
from lutris.gui.config.runner_box import RunnerBox
from lutris.gui.config.services_box import ServicesBox
from lutris.gui.dialogs import display_error
from lutris.gui.dialogs.runner_install import RunnerInstallDialog
from lutris.gui.widgets.utils import get_widget_children, pick_stock_icon
from lutris.installer.interpreter import ScriptInterpreter
from lutris.runners import InvalidRunnerError
from lutris.services import SERVICES
from lutris.services.base import (
    SERVICE_GAMES_LOADED,
    SERVICE_GAMES_LOADING,
    SERVICE_LOGIN,
    SERVICE_LOGOUT,
    AuthTokenExpiredError,
)
from lutris.util.jobs import schedule_at_idle
from lutris.util.library_sync import LOCAL_LIBRARY_SYNCED, LOCAL_LIBRARY_SYNCING
from lutris.util.log import logger
from lutris.util.strings import get_natural_sort_key

TYPE = 0
SLUG = 1
ICON = 2
LABEL = 3
GAMECOUNT = 4

SERVICE_INDICES = {name: index for index, name in enumerate(SERVICES.keys())}


class SidebarRow(Gtk.ListBoxRow):
    """A row in the sidebar containing possible action buttons"""

    MARGIN = 9
    SPACING = 6

    def __init__(self, id_, type_, name, icon, application=None):
        """Initialize the row

        Parameters:
            id_: identifier of the row
            type: type of row to display (still used?)
            name (str): Text displayed on the row
            icon (GtkImage): icon displayed next to the label
            application (GtkApplication): reference to the running application
        """
        super().__init__()
        self.application = application
        self.type = type_
        self.id = id_
        self.name = name
        self.is_updating = False
        self.buttons = {}
        self.box = Gtk.Box(spacing=self.SPACING, margin_start=self.MARGIN, margin_end=self.MARGIN)
        self.connect("realize", self.on_realize)
        self.add(self.box)

        if not icon:
            icon = Gtk.Box(spacing=self.SPACING, margin_start=self.MARGIN, margin_end=self.MARGIN)
        self.box.add(icon)
        label = Gtk.Label(
            label=name,
            halign=Gtk.Align.START,
            hexpand=True,
            margin_top=self.SPACING,
            margin_bottom=self.SPACING,
            ellipsize=Pango.EllipsizeMode.END,
        )
        self.box.pack_start(label, True, True, 0)
        self.btn_box = Gtk.Box(spacing=3, no_show_all=True, valign=Gtk.Align.CENTER, homogeneous=True)
        self.box.pack_end(self.btn_box, False, False, 0)
        self.spinner = Gtk.Spinner(visible=False, no_show_all=True)
        self.box.pack_end(self.spinner, False, False, 0)

    @property
    def sort_key(self) -> Union[int, str]:
        """An index indicate the place this row has within its type. The id is used
        as a tie-breaker. Can be int or str depending on row type."""
        return 0

    def get_actions(self):
        return []

    def is_row_active(self):
        """Return true if the row is hovered or is the one selected"""
        flags = self.get_state_flags()
        # Naming things sure is hard... But "prelight" instead of "hover"? Come on...
        return flags & Gtk.StateFlags.PRELIGHT or flags & Gtk.StateFlags.SELECTED

    def do_state_flags_changed(self, previous_flags):  # pylint: disable=arguments-differ
        if self.id:
            self.update_buttons()
        Gtk.ListBoxRow.do_state_flags_changed(self, previous_flags)

    def update_buttons(self):
        if self.is_updating:
            self.btn_box.hide()
            self.spinner.show()
            self.spinner.start()
            return
        self.spinner.stop()
        self.spinner.hide()
        if self.is_row_active():
            self.btn_box.show()
        elif self.btn_box.get_visible():
            self.btn_box.hide()

    def create_button_box(self):
        """Adds buttons in the button box based on the row's actions"""
        for child in self.btn_box.get_children():
            child.destroy()
        for icon_name, text, clicked, key in self.get_actions():
            icon_name = pick_stock_icon(icon_name, fallback_name="preferences-system-symbolic")
            btn = Gtk.Button(tooltip_text=text, relief=Gtk.ReliefStyle.NONE, visible=True)
            image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.MENU)
            image.show()
            btn.add(image)
            btn.connect("clicked", clicked)
            self.buttons[key] = btn
            self.btn_box.add(btn)

    def on_realize(self, widget):
        self.create_button_box()


class ServiceSidebarRow(SidebarRow):
    def __init__(self, service):
        super().__init__(service.id, "service", service.name, LutrisSidebar.get_sidebar_icon(service.icon))
        self.service = service

    @property
    def sort_key(self):
        return SERVICE_INDICES[self.id]

    def get_actions(self):
        """Return the definition of buttons to be added to the row"""
        displayed_buttons = []
        if self.service.is_launchable():
            displayed_buttons.append(("media-playback-start-symbolic", _("Run"), self.on_service_run, "run"))
        displayed_buttons.append(("view-refresh-symbolic", _("Reload"), self.on_refresh_clicked, "refresh"))
        return displayed_buttons

    def on_service_run(self, button):
        """Run a launcher associated with a service"""
        self.service.run(self.get_toplevel())

    def on_refresh_clicked(self, button):
        """Reload the service games"""
        button.set_sensitive(False)
        if self.service.online and not self.service.is_connected():
            self.service.logout()
            self.service.login(parent=self.get_toplevel())  # login will trigger reload if successful
            return
        self.service.start_reload(self.service_reloaded_cb)

    def service_reloaded_cb(self, error):
        if error:
            if isinstance(error, AuthTokenExpiredError):
                self.service.logout()
                self.service.login(parent=self.get_toplevel())  # login will trigger reload if successful
            else:
                display_error(error, parent=self.get_toplevel())
        schedule_at_idle(self.enable_refresh_button, delay_seconds=2.0)

    def enable_refresh_button(self) -> None:
        self.buttons["refresh"].set_sensitive(True)


class OnlineServiceSidebarRow(ServiceSidebarRow):
    def get_buttons(self):
        return {
            "run": (("media-playback-start-symbolic", _("Run"), self.on_service_run, "run")),
            "refresh": ("view-refresh-symbolic", _("Reload"), self.on_refresh_clicked, "refresh"),
            "disconnect": ("system-log-out-symbolic", _("Disconnect"), self.on_connect_clicked, "disconnect"),
            "connect": ("avatar-default-symbolic", _("Connect"), self.on_connect_clicked, "connect"),
        }

    def get_actions(self):
        buttons = self.get_buttons()
        displayed_buttons = []
        if self.service.is_launchable():
            displayed_buttons.append(buttons["run"])
        if self.service.is_authenticated():
            displayed_buttons += [buttons["refresh"], buttons["disconnect"]]
        else:
            displayed_buttons += [buttons["connect"]]
        return displayed_buttons

    def on_connect_clicked(self, button):
        button.set_sensitive(False)
        if self.service.is_authenticated():
            self.service.logout()
        else:
            self.service.login(parent=self.get_toplevel())
        self.create_button_box()


class RunnerSidebarRow(SidebarRow):
    def get_actions(self):
        """Return the definition of buttons to be added to the row"""
        if not self.id:
            return []
        entries = []

        # Creation is delayed because only installed runners can be imported
        # and all visible boxes should be installed.
        try:
            runner = self.get_runner()
        except InvalidRunnerError:
            return entries

        if runner.multiple_versions:
            entries.append(
                ("system-software-install-symbolic", _("Manage Versions"), self.on_manage_versions, "manage-versions")
            )
        if runner.runnable_alone:
            entries.append(("media-playback-start-symbolic", _("Run"), self.on_run_runner, "run"))
        entries.append(("emblem-system-symbolic", _("Configure"), self.on_configure_runner, "configure"))
        return entries

    def get_runner(self):
        return runners.import_runner(self.id)()

    def on_run_runner(self, *_args):
        """Runs the runner without no game."""
        runner = self.get_runner()
        runner.run(self.get_toplevel())

    def on_configure_runner(self, *_args):
        """Show runner configuration"""
        runner = self.get_runner()
        self.application.show_window(RunnerConfigDialog, runner=runner, parent=self.get_toplevel())

    def on_manage_versions(self, *_args):
        """Manage runner versions"""
        runner = self.get_runner()
        dlg_title = _("Manage %s versions") % runner.human_name
        self.application.show_window(RunnerInstallDialog, title=dlg_title, runner=runner, parent=self.get_toplevel())


class CategorySidebarRow(SidebarRow):
    def __init__(self, category, application):
        super().__init__(
            category["name"],
            "user_category",
            category["name"],
            LutrisSidebar.get_sidebar_icon("folder-symbolic"),
            application=application,
        )
        self.category = category

        self._sort_name = locale.strxfrm(category["name"])

    @property
    def sort_key(self):
        return get_natural_sort_key(self.name)

    def get_actions(self):
        """Return the definition of buttons to be added to the row"""
        return [("applications-system-symbolic", _("Edit Games"), self.on_category_clicked, "manage-category-games")]

    def on_category_clicked(self, button):
        category = categories_db.get_category_by_id(self.category["id"]) or self.category
        self.application.show_window(EditCategoryGamesDialog, category=category, parent=self.get_toplevel())
        return True

    def __lt__(self, other):
        if not isinstance(other, CategorySidebarRow):
            raise ValueError("Cannot compare %s to %s" % (self.__class__.__name__, other.__class__.__name__))

        return self._sort_name < other._sort_name

    def __gt__(self, other):
        if not isinstance(other, CategorySidebarRow):
            raise ValueError("Cannot compare %s to %s" % (self.__class__.__name__, other.__class__.__name__))

        return self._sort_name > other._sort_name


class SavedSearchSidebarRow(SidebarRow):
    def __init__(self, saved_search: saved_search_db.SavedSearch, application):
        super().__init__(
            saved_search.name,
            "saved_search",
            saved_search.name,
            LutrisSidebar.get_sidebar_icon("folder-saved-search-symbolic"),
            application=application,
        )
        self.saved_search = saved_search

        self._sort_name = locale.strxfrm(saved_search.name)

    @property
    def sort_key(self):
        return get_natural_sort_key(self.name)

    def get_actions(self):
        """Return the definition of buttons to be added to the row"""
        return [
            ("applications-system-symbolic", _("Edit Games"), self.on_saved_search_clicked, "manage-category-games")
        ]

    def on_saved_search_clicked(self, button):
        saved_search = saved_search_db.get_saved_search_by_id(self.saved_search.saved_search_id) or self.saved_search
        self.application.show_window(EditSavedSearchDialog, saved_search=saved_search, parent=self.get_toplevel())
        return True

    def __lt__(self, other):
        if not isinstance(other, SavedSearchSidebarRow):
            raise ValueError("Cannot compare %s to %s" % (self.__class__.__name__, other.__class__.__name__))

        return self._sort_name < other._sort_name

    def __gt__(self, other):
        if not isinstance(other, SavedSearchSidebarRow):
            raise ValueError("Cannot compare %s to %s" % (self.__class__.__name__, other.__class__.__name__))

        return self._sort_name > other._sort_name


class SidebarHeader(Gtk.ListBoxRow):
    """Header shown on top of each sidebar section, as a selectable row"""

    def __init__(
        self,
        name: str,
        header_index: int,
        section_id: str,
        collapsible: bool = True,
        on_toggle: Optional[Callable[[str, bool], None]] = None,
    ) -> None:
        super().__init__()
        self.set_selectable(False)
        self.set_activatable(False)
        self.header_index: int = header_index
        self.section_id: str = section_id
        self.type: str = "header"
        self.id: str = section_id
        self.collapsible: bool = collapsible
        self.on_toggle: Optional[Callable[[str, bool], None]] = on_toggle
        self._collapsed: bool = False

        outer_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        label = Gtk.Label(
            halign=Gtk.Align.START,
            hexpand=True,
            use_markup=True,
            label="<b>{}</b>".format(name),
        )

        self.arrow = Gtk.Image.new_from_icon_name("pan-down-symbolic", Gtk.IconSize.MENU)
        self.arrow.set_margin_end(6)

        box = Gtk.Box(margin_start=9, margin_top=6, margin_bottom=6, margin_end=9)
        if collapsible:
            box.pack_start(self.arrow, False, False, 0)
        box.pack_start(label, True, True, 0)

        if collapsible:
            event_box = Gtk.EventBox()
            event_box.add(box)
            event_box.connect("button-press-event", self._on_click)
            event_box.connect("enter-notify-event", self._on_enter)
            event_box.connect("leave-notify-event", self._on_leave)
            outer_box.add(event_box)
        else:
            outer_box.add(box)

        outer_box.add(Gtk.Separator())
        self.add(outer_box)
        self.show_all()

    @property
    def collapsed(self) -> bool:
        return self._collapsed

    @collapsed.setter
    def collapsed(self, value: bool) -> None:
        self._collapsed = value
        if self.collapsible:
            icon_name = "pan-end-symbolic" if value else "pan-down-symbolic"
            self.arrow.set_from_icon_name(icon_name, Gtk.IconSize.MENU)

    def _on_click(self, widget: Gtk.EventBox, event: Gdk.EventButton) -> bool:
        if event.button == 1:  # Left click
            self.collapsed = not self.collapsed
            if self.on_toggle:
                self.on_toggle(self.section_id, self.collapsed)
        return True

    def _on_enter(self, widget: Gtk.EventBox, event: Gdk.EventCrossing) -> None:
        window = widget.get_window()
        if window:
            window.set_cursor(Gdk.Cursor.new_from_name(widget.get_display(), "pointer"))

    def _on_leave(self, widget: Gtk.EventBox, event: Gdk.EventCrossing) -> None:
        window = widget.get_window()
        if window:
            window.set_cursor(None)

    @property
    def sort_key(self) -> Union[int, str]:
        # Headers sort before all rows in their section (not used directly in sorting)
        return 0


class DummyRow:
    """Dummy class for rows that may not be initialized."""

    def show(self):
        """Dummy method for showing the row"""

    def hide(self):
        """Dummy method for hiding the row"""


class LutrisSidebar(Gtk.ListBox):
    __gtype_name__ = "LutrisSidebar"

    def __init__(self, application):
        super().__init__()
        self.set_size_request(200, -1)
        self.application = application
        self.previous_category = None
        self.get_style_context().add_class("lutris-sidebar")

        # Empty values until LutrisWindow explicitly initializes the rows
        # at the right time.
        self.installed_runners = []
        self.runner_visibility_cache = {}
        self.used_categories = set()
        self.saved_searches = set()
        self.active_services = {}
        self.active_platforms = []
        self.service_rows = {}
        self.runner_rows = {}
        self.platform_rows = {}

        self.category_rows = {}
        self.saved_search_rows = {}
        # A dummy objects that allows inspecting why/when we have a show() call on the object.
        self.games_row = DummyRow()
        self.running_row = DummyRow()
        self.hidden_row = DummyRow()
        self.missing_row = DummyRow()

        # Load collapsed sections state from settings
        self.collapsed_sections = self._load_collapsed_sections()

        self.row_headers = {
            "library": SidebarHeader(_("Library"), header_index=0, section_id="library", collapsible=False),
            "user_category": SidebarHeader(
                _("Categories"), header_index=1, section_id="user_category", on_toggle=self._on_section_toggle
            ),
            "saved_search": SidebarHeader(
                _("Saved Searches"), header_index=2, section_id="saved_search", on_toggle=self._on_section_toggle
            ),
            "service": SidebarHeader(
                _("Sources"), header_index=3, section_id="service", on_toggle=self._on_section_toggle
            ),
            "runner": SidebarHeader(
                _("Runners"), header_index=4, section_id="runner", on_toggle=self._on_section_toggle
            ),
            "platform": SidebarHeader(
                _("Platforms"), header_index=5, section_id="platform", on_toggle=self._on_section_toggle
            ),
        }

        # Restore collapsed state for each header
        for section_id, header in self.row_headers.items():
            if section_id in self.collapsed_sections:
                header.collapsed = True
        GObject.add_emission_hook(RunnerBox, "runner-installed", self.update_rows)
        GObject.add_emission_hook(RunnerBox, "runner-removed", self.update_rows)
        GObject.add_emission_hook(RunnerConfigDialog, "runner-updated", self.update_runner_rows)
        GObject.add_emission_hook(ScriptInterpreter, "runners-installed", self.update_rows)
        GObject.add_emission_hook(ServicesBox, "services-changed", self.update_rows)
        GAME_START.register(self.on_game_start)
        GAME_STOPPED.register(self.on_game_stopped)
        GAME_UPDATED.register(self.update_rows)
        CATEGORIES_UPDATED.register(self.update_rows)
        SAVED_SEARCHES_UPDATED.register(self.update_rows)
        SERVICE_LOGIN.register(self.on_service_auth_changed)
        SERVICE_LOGOUT.register(self.on_service_auth_changed)
        SERVICE_GAMES_LOADING.register(self.on_service_games_loading)
        SERVICE_GAMES_LOADED.register(self.on_service_games_loaded)
        LOCAL_LIBRARY_SYNCING.register(self.on_local_library_syncing)
        LOCAL_LIBRARY_SYNCED.register(self.on_local_library_synced)
        self.set_filter_func(self._filter_func)
        self.show_all()

    @staticmethod
    def get_sidebar_icon(icon_name: str, fallback_icon_names: List[str] = None) -> Gtk.Image:
        candidate_names = [icon_name] + (fallback_icon_names or [])
        icon_name = pick_stock_icon(candidate_names, fallback_name="package-x-generic-symbolic")
        icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.MENU)

        # We can wind up with an icon of the wrong size, if that's what is
        # available. So we'll fix that.
        icon_size = Gtk.IconSize.lookup(Gtk.IconSize.MENU)
        if icon_size[0]:
            icon.set_pixel_size(icon_size[2])

        return icon

    def _load_collapsed_sections(self) -> Set[str]:
        """Load collapsed sections state from settings."""
        collapsed_str = settings.read_setting("sidebar_collapsed_sections", default="")
        if collapsed_str:
            return set(collapsed_str.split(","))
        return set()

    def _save_collapsed_sections(self) -> None:
        """Save collapsed sections state to settings."""
        collapsed_str = ",".join(sorted(self.collapsed_sections))
        settings.write_setting("sidebar_collapsed_sections", collapsed_str)

    def _on_section_toggle(self, section_id: str, collapsed: bool) -> None:
        """Handle section collapse/expand toggle."""
        if collapsed:
            self.collapsed_sections.add(section_id)
        else:
            self.collapsed_sections.discard(section_id)
        self._save_collapsed_sections()
        self.invalidate_filter()

    def is_section_collapsed(self, section_type: str) -> bool:
        """Check if a section is collapsed."""
        return section_type in self.collapsed_sections

    def initialize_rows(self):
        """
        Select the initial row; this triggers the initialization of the game view,
        so we must do this even if this sidebar is never realized, but only after
        the sidebar's signals are connected.
        """

        # Add the Library header first
        self.add(self.row_headers["library"])

        # Create the basic rows that are not data dependant

        self.games_row = SidebarRow(
            "all",
            "category",
            _("Games"),
            self.get_sidebar_icon("applications-games-symbolic"),
        )
        self.add(self.games_row)

        self.add(
            SidebarRow(
                "recent",
                "dynamic_category",
                _("Recent"),
                self.get_sidebar_icon("document-open-recent-symbolic"),
            )
        )

        self.add(
            SidebarRow(
                "favorite",
                "category",
                _("Favorites"),
                self.get_sidebar_icon("favorite-symbolic"),
            )
        )

        self.add(
            SidebarRow(
                ".uncategorized",
                "dynamic_category",
                _("Uncategorized"),
                self.get_sidebar_icon("tag-symbolic", ["poi-marker", "favorite-symbolic"]),
            )
        )

        self.hidden_row = SidebarRow(
            ".hidden",
            "category",
            _("Hidden"),
            self.get_sidebar_icon("action-unavailable-symbolic"),
        )
        self.add(self.hidden_row)

        self.missing_row = SidebarRow(
            "missing",
            "dynamic_category",
            _("Missing"),
            self.get_sidebar_icon("dialog-warning-symbolic"),
        )
        self.add(self.missing_row)

        self.running_row = SidebarRow(
            "running",
            "dynamic_category",
            _("Running"),
            self.get_sidebar_icon("media-playback-start-symbolic"),
        )
        # I wanted this to be on top but it really messes with the headers when showing/hiding the row.
        self.add(self.running_row)
        self.show_all()
        self.hidden_row.hide()
        self.missing_row.hide()
        self.running_row.hide()

        # Create the dynamic rows that are initially needed
        self.update_rows()

    @property
    def selected_category(self):
        """The selected sidebar row, as a tuple of category type and category value,
        like ('service', 'lutris')."""
        row = self.get_selected_row()
        return (row.type, row.id) if row else ("category", "all")

    @selected_category.setter
    def selected_category(self, value):
        """Selects the row for the category indicated by a category tuple,
        like ('service', 'lutris')"""
        self.previous_category = self.selected_category
        selected_row_type, selected_row_id = value or ("category", "all")
        children = get_widget_children(self, SidebarRow)
        for row in children:
            if row.type == selected_row_type and row.id == selected_row_id:
                if row.get_visible():
                    self.select_row(row)
                    return

                break

        for row in children:
            if row.get_visible():
                self.select_row(row)
                return

    def _filter_func(self, row):
        def is_runner_visible(runner_name):
            if runner_name not in self.runner_visibility_cache:
                runner_config = LutrisConfig(runner_slug=row.id, options_supported={"visible_in_side_panel"})
                self.runner_visibility_cache[runner_name] = runner_config.runner_config.get(
                    "visible_in_side_panel", True
                )
            return self.runner_visibility_cache[runner_name]

        # Header rows: show if their section has content
        if row.type == "header":
            section_id = row.section_id
            if section_id == "library":
                return True
            if section_id == "user_category":
                return bool(self.used_categories)
            if section_id == "saved_search":
                return bool(self.saved_searches)
            if section_id == "service":
                return bool(self.active_services)
            if section_id == "runner":
                return bool(self.installed_runners)
            if section_id == "platform":
                return bool(self.active_platforms)
            return True

        if not row or not row.id or row.type in ("category", "dynamic_category"):
            return True

        # Hide rows if their section is collapsed
        if self.is_section_collapsed(row.type):
            return False

        if row.type == "runner":
            if row.id is None:
                return True  # 'All'
            return row.id in self.installed_runners and is_runner_visible(row.id)

        if row.type == "user_category":
            allowed_ids = self.used_categories
        elif row.type == "saved_search":
            allowed_ids = self.saved_searches
        elif row.type == "service":
            allowed_ids = self.active_services
        else:
            allowed_ids = self.active_platforms

        return row.id in allowed_ids

    def update_runner_rows(self, *_args):
        self.runner_visibility_cache.clear()
        self.update_rows()
        return True

    def update_rows(self, *_args):
        """Generates any missing rows that are now needed, and re-evaluate the filter to hide
        any no longer needed. GTK has a lot of trouble dynamically updating and re-arranging
        rows, so this will have to do. This keeps the total row count down reasonably well."""

        # Type alias for sort key: (header_index, is_header, sort_key, id)
        SortKey = Tuple[int, bool, Union[int, str], str]

        def get_sort_key(row: Gtk.ListBoxRow) -> SortKey:
            """Returns a key used to sort the rows. This keeps rows for a header
            together, and rows in a hopefully reasonable order as we insert them.
            Returns (header_index, is_header, sort_key, id) where is_header ensures
            headers sort before content rows."""
            if isinstance(row, SidebarHeader):
                # Headers sort at the start of their section (is_header=False sorts before True)
                return row.header_index, False, 0, ""
            if isinstance(row, SidebarRow):
                header_row = self.row_headers.get(row.type) if row.type else None
                header_index = header_row.header_index if header_row else 0
                return header_index, True, row.sort_key, row.id
            # Fallback for any other row types
            return 0, True, 0, ""

        def insert_row(row: Gtk.ListBoxRow) -> None:
            """Find the best place to insert the row, to maintain order, and inserts it there."""
            index = 0
            seq = get_sort_key(row)
            while True:
                r = self.get_row_at_index(index)
                if not r or get_sort_key(r) > seq:
                    break
                index += 1

            row.show_all()
            self.insert(row, index)

        categories_db.remove_unused_categories()
        categories = [c for c in categories_db.get_categories() if not categories_db.is_reserved_category(c["name"])]
        saved_searches = saved_search_db.get_saved_searches()

        self.used_categories = {c["name"] for c in categories}
        self.saved_searches = {s.name for s in saved_searches}
        self.active_services = services.get_enabled_services()
        self.installed_runners = [runner.name for runner in runners.get_installed()]
        self.active_platforms = games_db.get_used_platforms()

        # Add section headers if not already in the sidebar
        for section_id, header in self.row_headers.items():
            if section_id != "library" and header.get_parent() is None:
                insert_row(header)

        for service_name, service_class in self.active_services.items():
            if service_name not in self.service_rows:
                try:
                    service = service_class()
                    row_class = OnlineServiceSidebarRow if service.online else ServiceSidebarRow
                    service_row = row_class(service)
                    insert_row(service_row)
                    self.service_rows[service_name] = service_row
                except Exception as ex:
                    logger.exception("Sidebar row for '%s' could not be loaded: %s", service_name, ex)

        for runner_name in self.installed_runners:
            if runner_name not in self.runner_rows:
                try:
                    icon_name = runner_name.lower().replace(" ", "") + "-symbolic"
                    runner = runners.import_runner(runner_name)()
                    runner_row = RunnerSidebarRow(
                        runner_name,
                        "runner",
                        runner.human_name,
                        self.get_sidebar_icon(icon_name),
                        application=self.application,
                    )
                    insert_row(runner_row)
                    self.runner_rows[runner_name] = runner_row
                except Exception as ex:
                    logger.exception("Sidebar row for '%s' could not be loaded: %s", service_name, ex)

        for platform in self.active_platforms:
            if platform not in self.platform_rows:
                icon_name = platform.lower().replace(" ", "").replace("/", "_") + "-symbolic"
                platform_row = SidebarRow(
                    platform, "platform", platform, self.get_sidebar_icon(icon_name), application=self.application
                )
                self.platform_rows[platform] = platform_row
                insert_row(platform_row)

        # Remove stale category rows that no longer exist in the database
        stale_categories = set(self.category_rows.keys()) - self.used_categories
        for stale_name in stale_categories:
            stale_row = self.category_rows.pop(stale_name)
            stale_row.destroy()

        for category in categories:
            if category["name"] not in self.category_rows:
                new_category_row = CategorySidebarRow(category, application=self.application)
                self.category_rows[category["name"]] = new_category_row
                insert_row(new_category_row)

        for saved_search in saved_searches:
            if saved_search.name not in self.saved_search_rows:
                new_saved_search_row = SavedSearchSidebarRow(saved_search, application=self.application)
                self.saved_search_rows[saved_search.name] = new_saved_search_row
                insert_row(new_saved_search_row)

        self.invalidate_filter()
        return True

    def on_game_start(self, _game: Game) -> None:
        """Show the "running" section when a game start"""
        self.running_row.show()

    def on_game_stopped(self, _game: Game) -> None:
        """Hide the "running" section when no games are running"""
        if not self.application.has_running_games:
            self.running_row.hide()

            if self.get_selected_row() == self.running_row:
                self.select_row(get_widget_children(self, SidebarRow)[0])

    def on_service_auth_changed(self, service):
        logger.debug("Service %s auth changed", service.id)
        if service.id in self.service_rows:
            self.service_rows[service.id].create_button_box()
            self.service_rows[service.id].update_buttons()
        else:
            logger.warning("Service %s is not found", service.id)
        return True

    def on_service_games_loading(self, service):
        logger.debug("Service %s games loading", service.id)
        if service.id in self.service_rows:
            self.service_rows[service.id].is_updating = True
            self.service_rows[service.id].update_buttons()
        else:
            logger.warning("Service %s is not found", service.id)
        return True

    def on_service_games_loaded(self, service):
        logger.debug("Service %s games loaded", service.id)
        if service.id in self.service_rows:
            self.service_rows[service.id].is_updating = False
            self.service_rows[service.id].update_buttons()
        else:
            logger.warning("Service %s is not found", service.id)
        return True

    def on_local_library_syncing(self):
        self.games_row.is_updating = True
        self.games_row.update_buttons()

    def on_local_library_synced(self):
        self.games_row.is_updating = False
        self.games_row.update_buttons()



"""isort:skip_file"""

import os
from gettext import gettext as _
from typing import TYPE_CHECKING

import gi

if TYPE_CHECKING:
    from lutris.services.base import OnlineService

try:
    gi.require_version("WebKit2", "4.1")
except ValueError:
    gi.require_version("WebKit2", "4.0")
from gi.repository import WebKit2  # type: ignore[attr-defined]

from lutris.config import LutrisConfig
from lutris.gui.dialogs import ModalDialog
from lutris.util.log import logger


class WebConnectDialog(ModalDialog):
    """Login form for external services"""

    def __init__(self, service: "OnlineService", parent=None):
        if service.is_login_in_progress:
            # If the previous login was abandoned, remove any
            # credentials that may have been left over for a clean start-
            # this helps if the service thinks we are logged in (ie, credenial
            # cookiers remain), but Lutris does not.
            service.wipe_game_cache()

        service.is_login_in_progress = True

        self.context: WebKit2.WebContext = WebKit2.WebContext.new()

        # Set locale
        # Locale fallback routine:
        # Lutris locale -> System environment locale -> US English
        webview_locales = ["en_US"]
        lutris_config = LutrisConfig()
        environment_locale_lang = os.environ.get("LANG")
        if environment_locale_lang:
            webview_locales = [environment_locale_lang.split(".")[0]] + webview_locales
        lutris_locale = lutris_config.system_config.get("locale")
        if lutris_locale:
            webview_locales = [lutris_locale.split(".")[0]] + webview_locales
        logger.debug(
            f"Webview locale fallback order: "
            f"[Lutris locale]: '{lutris_locale}' -> "
            f"[env: LANG]: '{environment_locale_lang}' -> "
            f"[Default]: '{webview_locales[-1]}'"
        )
        self.context.set_preferred_languages(webview_locales)

        if "http_proxy" in os.environ:
            proxy = WebKit2.NetworkProxySettings.new(os.environ["http_proxy"])
            self.context.set_network_proxy_settings(WebKit2.NetworkProxyMode.CUSTOM, proxy)
        WebKit2.CookieManager.set_persistent_storage(
            self.context.get_cookie_manager(),
            service.cookies_path,
            WebKit2.CookiePersistentStorage(0),
        )
        self.service = service

        super().__init__(title=service.name, parent=parent)

        content_area = self.get_content_area()

        self.set_default_size(self.service.login_window_width, self.service.login_window_height)

        self.webview = WebKit2.WebView.new_with_context(self.context)
        self.webview.load_uri(service.login_url)
        self.webview.connect("load-changed", self.on_navigation)
        self.webview.connect("create", self.on_webview_popup)
        self.webview.connect("decide-policy", self.on_decide_policy)
        content_area.set_border_width(0)
        content_area.pack_start(self.webview, True, True, 0)

        webkit_settings = self.webview.get_settings()

        # Set a User Agent
        webkit_settings.set_user_agent(service.login_user_agent)

        # Allow popups (Doesn't work...)
        webkit_settings.set_allow_modal_dialogs(True)

        # Enable developer options for troubleshooting (Can be disabled in
        # releases)
        # webkit_settings.set_enable_write_console_messages_to_stdout(True)
        # webkit_settings.set_javascript_can_open_windows_automatically(True)
        webkit_settings.set_enable_developer_extras(True)
        webkit_settings.set_enable_webgl(False)
        # self.enable_inspector()
        self.show_all()

    def enable_inspector(self):
        """If you want a full blown Webkit inspector, call this"""
        # WARNING: For some reason this doesn't work as intended.
        # The inspector shows ups but it's impossible to interact with it
        # All inputs are blocked by the the webkit dialog.
        inspector = self.webview.get_inspector()
        inspector.show()

    def on_decide_policy(self, webview, decision, decision_type):
        if decision_type == WebKit2.PolicyDecisionType.NAVIGATION_ACTION:
            decision.use()
        return True

    def on_navigation(self, widget, load_event):
        if load_event == WebKit2.LoadEvent.FINISHED:
            url = widget.get_uri()
            if url in self.service.scripts:
                script = self.service.scripts[url]
                widget.run_javascript(script, None, None)
                return True
            if any(url.startswith(r) for r in self.service.redirect_uris):
                if self.service.requires_login_page:
                    resource = widget.get_main_resource()
                    resource.get_data(None, self._get_response_data_finish, None)
                else:
                    self.service.is_login_in_progress = False
                    self.service.login_callback(url)
                    self.destroy()
        return True

    def _get_response_data_finish(self, resource, result, user_data=None):
        html_response = resource.get_data_finish(result)
        self.service.is_login_in_progress = False
        self.service.login_callback(html_response)
        self.destroy()

    def on_webview_popup(self, widget, navigation_action):
        """Handles web popups created by this dialog's webview"""
        uri = navigation_action.get_request().get_uri()
        view = WebKit2.WebView.new_with_related_view(widget)
        popup_dialog = WebPopupDialog(view, uri, parent=self)
        popup_dialog.show()
        return view


class WebPopupDialog(ModalDialog):
    """Dialog for handling web popups"""

    def __init__(self, webview, uri, parent=None):
        # pylint: disable=no-member
        self.parent = parent
        super().__init__(title=_("Loading..."), parent=parent)
        self.webview = webview
        self.webview.connect("ready-to-show", self.on_ready_webview)
        self.webview.connect("notify::title", self.on_available_webview_title)
        self.webview.connect("create", self.on_new_webview_popup)
        self.webview.connect("close", self.on_webview_close)
        self.webview.load_uri(uri)
        self.vbox.pack_start(self.webview, True, True, 0)
        self.vbox.set_border_width(0)
        self.set_default_size(390, 500)

    def on_ready_webview(self, webview):
        self.show_all()

    def on_available_webview_title(self, webview, gparamstring):
        self.set_title(webview.get_title())

    def on_new_webview_popup(self, webview, navigation_action):
        """Handles web popups created by this dialog's webview"""
        uri = navigation_action.get_request().get_uri()
        view = WebKit2.WebView.new_with_related_view(webview)
        view.load_uri(uri)
        dialog = WebPopupDialog(view, uri, parent=self)
        dialog.show()
        return view

    def on_webview_close(self, webview):
        self.close()
        return True



"""AppIndicator/AyatanaAppIndicator based tray icon"""

from gettext import gettext as _

import gi
from gi.repository import Gtk

from lutris.database import categories
from lutris.database.games import get_games
from lutris.game import Game
from lutris.util.display import is_display_x11

try:
    gi.require_version("AppIndicator3", "0.1")
    from gi.repository import AppIndicator3 as AppIndicator

    APP_INDICATOR_SUPPORTED = True
except (ImportError, ValueError):
    try:
        gi.require_version("AyatanaAppIndicator3", "0.1")
        from gi.repository import AyatanaAppIndicator3 as AppIndicator

        APP_INDICATOR_SUPPORTED = True
    except (ImportError, ValueError):
        APP_INDICATOR_SUPPORTED = False


def supports_status_icon() -> bool:
    return bool(APP_INDICATOR_SUPPORTED or is_display_x11())


class LutrisStatusIcon:
    """This is a proxy for the status icon, which can be an AppIndicator or a Gtk.StatusIcon. Or if
    neither is supported, it can be a null object that silently does nothing."""

    def __init__(self, application):
        self.application = application
        self.indicator = None
        self.tray_icon = None
        self.menu = None
        self.present_menu = None

        if supports_status_icon():
            self.menu = self._get_menu()
            if APP_INDICATOR_SUPPORTED:
                self.indicator = AppIndicator.Indicator.new(
                    "net.lutris.Lutris", "net.lutris.Lutris", AppIndicator.IndicatorCategory.APPLICATION_STATUS
                )
                self.indicator.set_menu(self.menu)
            else:
                self.tray_icon = self._get_tray_icon()
                self.tray_icon.connect("activate", self.on_activate)
                self.tray_icon.connect("popup-menu", self.on_menu_popup)

            self.set_visible(True)

    def is_visible(self):
        """Whether the icon is visible"""
        if self.indicator:
            return self.indicator.get_status() != AppIndicator.IndicatorStatus.PASSIVE

        if self.tray_icon:
            return self.tray_icon.get_visible()

        return False

    def set_visible(self, value):
        """Set the visibility of the icon"""
        if self.indicator:
            if value:
                visible = AppIndicator.IndicatorStatus.ACTIVE
            else:
                visible = AppIndicator.IndicatorStatus.PASSIVE
            self.indicator.set_status(visible)
        elif self.tray_icon:
            self.tray_icon.set_visible(value)

    def _get_menu(self):
        """Instantiates the menu attached to the tray icon"""
        menu = Gtk.Menu()
        installed_games = self._get_installed_games()
        number_of_games_in_menu = 10
        for game in installed_games[:number_of_games_in_menu]:
            menu.append(self._make_menu_item_for_game(game))
        menu.append(Gtk.SeparatorMenuItem())

        self.present_menu = Gtk.ImageMenuItem()
        self.present_menu.set_image(Gtk.Image.new_from_icon_name("net.lutris.Lutris", Gtk.IconSize.MENU))
        self.present_menu.set_label(_("Show Lutris"))
        self.present_menu.connect("activate", self.on_activate)
        menu.append(self.present_menu)

        quit_menu = Gtk.MenuItem()
        quit_menu.set_label(_("Quit"))
        quit_menu.connect("activate", self.on_quit_application)
        menu.append(quit_menu)
        menu.show_all()
        return menu

    def _get_tray_icon(self):
        tray_icon = Gtk.StatusIcon()
        tray_icon.set_tooltip_text(_("Lutris"))
        tray_icon.set_visible(True)
        tray_icon.set_from_icon_name("net.lutris.Lutris")
        return tray_icon

    def update_present_menu(self):
        app_window = self.application.window
        if app_window and self.present_menu:
            if app_window.get_visible():
                self.present_menu.set_label(_("Hide Lutris"))
            else:
                self.present_menu.set_label(_("Show Lutris"))

    def on_activate(self, _status_icon, _event=None):
        """Callback to show or hide the window"""
        app_window = self.application.window
        if app_window.get_visible():
            # If the window has any transients, hiding it will hide them too
            # never to be shown again, which is broken. So we don't allow that.
            windows = Gtk.Window.list_toplevels()
            for w in windows:
                if w.get_visible() and w.get_transient_for() == app_window:
                    return

            app_window.hide()
        else:
            app_window.show()

    def on_menu_popup(self, _status_icon, button, time):
        """Callback to show the contextual menu"""
        self.menu.popup(None, None, None, None, button, time)

    def on_quit_application(self, _widget):
        """Callback to quit the program"""
        self.application.quit()

    def _make_menu_item_for_game(self, game):
        menu_item = Gtk.MenuItem()
        menu_item.set_label(game["name"])
        menu_item.connect("activate", self.on_game_selected, game["id"])
        return menu_item

    @staticmethod
    def _get_installed_games():
        """Adds installed games in order of last use"""
        installed_games = get_games(filters={"installed": 1})
        hidden_game_ids = categories.get_game_ids_for_categories([".hidden"])
        installed_games = [g for g in installed_games if g.get("id") not in hidden_game_ids]
        installed_games.sort(
            key=lambda game: max(game["lastplayed"] or 0, game["installed_at"] or 0),
            reverse=True,
        )
        return installed_games

    def on_game_selected(self, _widget, game_id):
        launch_ui_delegate = self.application.get_launch_ui_delegate()
        Game(game_id).launch(launch_ui_delegate)



"""Unused handler registration but since someone reports problems with URL
integration once in a while, it could prove itself useful."""

# Standard Library
import os
import sys

# Third Party Libraries
from gi.repository import Gio

# Lutris Modules
from lutris.util.log import logger


def register_url_handler():
    """Register the lutris: protocol to open with the application."""
    executable = os.path.abspath(sys.argv[0])
    base_key = "desktop.gnome.url-handlers.lutris"
    schema_directory = "/usr/share/glib-2.0/schemas/"
    schema_source = Gio.SettingsSchemaSource.new_from_directory(schema_directory, None, True)
    schema = schema_source.lookup(base_key, True)
    if schema:
        settings = Gio.Settings.new(base_key)
        settings.set_string("command", executable)
    else:
        logger.warning("Schema not installed, cannot register url-handler")



"""XDG shortcuts handling"""

import os
import shlex
import shutil
import stat
from textwrap import dedent

from gi.repository import GLib

from lutris.api import format_installer_url
from lutris.settings import CACHE_DIR
from lutris.util import system
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger


def get_lutris_executable():
    if LINUX_SYSTEM.is_flatpak():
        return "flatpak run net.lutris.Lutris"
    return "lutris"


def get_xdg_entry(directory):
    """Return the path for specific user folders"""
    special_dir = {
        "DESKTOP": GLib.UserDirectory.DIRECTORY_DESKTOP,
        "MUSIC": GLib.UserDirectory.DIRECTORY_MUSIC,
        "PICTURES": GLib.UserDirectory.DIRECTORY_PICTURES,
        "VIDEOS": GLib.UserDirectory.DIRECTORY_VIDEOS,
        "DOCUMENTS": GLib.UserDirectory.DIRECTORY_DOCUMENTS,
        "DOWNLOADS": GLib.UserDirectory.DIRECTORY_DOWNLOAD,
        "TEMPLATES": GLib.UserDirectory.DIRECTORY_TEMPLATES,
    }
    directory = directory.upper()
    if directory not in special_dir:
        raise ValueError(
            directory + " not supported. Only those folders are supported: " + ", ".join(special_dir.keys())
        )
    return GLib.get_user_special_dir(special_dir[directory])


def get_xdg_basename(game_slug, game_id, base_dir=None):
    """Return the filename for .desktop shortcuts"""
    if base_dir:
        # When base dir is provided, lookup possible combinations
        # and return the first match
        for path in [
            "net.lutris.{}-{}.desktop".format(game_slug, game_id),
            "{}-{}.desktop".format(game_slug, game_id),
            "{}.desktop".format(game_slug),
        ]:
            if system.path_exists(os.path.join(base_dir, path)):
                return path

    return "net.lutris.{}-{}.desktop".format(game_slug, game_id)


def create_launcher(game_slug, game_id, game_name, launch_config_name=None, desktop=False, menu=False):
    """Create a .desktop file."""
    desktop_dir = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DESKTOP)
    lutris_executable = get_lutris_executable()

    url = format_installer_url({"action": "rungameid", "game_slug": game_id, "launch_config_name": launch_config_name})

    # Quote URL for the shell but *also* quote %, which indicates a desktop file
    # field code in the Exec key.
    command = f"{lutris_executable} {shlex.quote(url)}".replace("%", "%%")

    try_exec = "" if LINUX_SYSTEM.is_flatpak() else "TryExec=lutris"

    launcher_content = dedent(
        """
        [Desktop Entry]
        Type=Application
        Name={}
        Icon={}
        Exec=env LUTRIS_SKIP_INIT=1 {}
        Categories=Game
        {}
        """.format(game_name, f"lutris_{game_slug}", command, try_exec)
    )

    launcher_filename = get_xdg_basename(game_slug, game_id)
    tmp_launcher_path = os.path.join(CACHE_DIR, launcher_filename)
    with open(tmp_launcher_path, "w", encoding="utf-8") as tmp_launcher:
        tmp_launcher.write(launcher_content)
        tmp_launcher.close()
    os.chmod(
        tmp_launcher_path,
        stat.S_IREAD | stat.S_IWRITE | stat.S_IRGRP | stat.S_IWGRP,
    )

    if desktop:
        os.makedirs(desktop_dir, exist_ok=True)
        launcher_path = os.path.join(desktop_dir, launcher_filename)
        logger.debug("Creating Desktop icon in %s", launcher_path)
        shutil.copy(tmp_launcher_path, launcher_path)
    if menu:
        user_dir = os.path.expanduser("~/.local/share") if LINUX_SYSTEM.is_flatpak() else GLib.get_user_data_dir()
        menu_path = os.path.join(user_dir, "applications")
        os.makedirs(menu_path, exist_ok=True)
        launcher_path = os.path.join(menu_path, launcher_filename)
        logger.debug("Creating menu launcher in %s", launcher_path)
        shutil.copy(tmp_launcher_path, launcher_path)
    os.remove(tmp_launcher_path)


def get_launcher_path(game_slug, game_id):
    """Return the path of a XDG game launcher.
    When legacy is set, it will return the old path with only the slug,
    otherwise it will return the path with slug + id
    """
    desktop_dir = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DESKTOP)

    return os.path.join(desktop_dir, get_xdg_basename(game_slug, game_id, base_dir=desktop_dir))


def get_menu_launcher_path(game_slug, game_id):
    """Return the path to a XDG menu launcher, prioritizing legacy paths if
    they exist
    """
    menu_dir = os.path.join(GLib.get_user_data_dir(), "applications")
    return os.path.join(menu_dir, get_xdg_basename(game_slug, game_id, base_dir=menu_dir))


def desktop_launcher_exists(game_slug, game_id):
    """Return True if there is an existing desktop icon for a game"""
    return system.path_exists(get_launcher_path(game_slug, game_id))


def menu_launcher_exists(game_slug, game_id):
    """Return True if there is an existing application menu entry for a game"""
    return system.path_exists(get_menu_launcher_path(game_slug, game_id))


def remove_launcher(game_slug, game_id, desktop=False, menu=False):
    """Remove existing .desktop file."""
    if desktop:
        launcher_path = get_launcher_path(game_slug, game_id)
        if system.path_exists(launcher_path):
            os.remove(launcher_path)

    if menu:
        menu_path = get_menu_launcher_path(game_slug, game_id)
        if system.path_exists(menu_path):
            os.remove(menu_path)



"""Timer module"""

# Standard Library
import time


class Timer:
    """Simple Timer class to time code"""

    def __init__(self):
        self._start = None
        self._end = None
        self.finished = False

    def start(self):
        """Starts the timer"""
        self._end = None
        self._start = time.monotonic()
        self.finished = False

    def end(self):
        """Ends the timer"""
        self._end = time.monotonic()
        self.finished = True

    @property
    def duration(self):
        """Return the total duration of the timer"""
        if not self._start:
            return 0

        if not self.finished:
            _duration = time.monotonic() - self._start
        else:
            _duration = self._end - self._start

        return _duration



"""Battle.net service"""

import json
import os
from gettext import gettext as _

from gi.repository import Gio

from lutris import settings
from lutris.config import LutrisConfig, write_game_config
from lutris.database.games import add_game, get_game_by_field
from lutris.database.services import ServiceGameCollection
from lutris.game import Game
from lutris.services.base import BaseService
from lutris.services.lutris import sync_media
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util.battlenet.definitions import ProductDbInfo
from lutris.util.log import logger

# Use pure-Python protobuf implementation for compatibility across all protobuf versions.
# The old-style generated pb2 file is incompatible with protobuf 4.x+ native implementation.
os.environ.setdefault("PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION", "python")

try:
    from lutris.util.battlenet.product_db_pb2 import ProductDb

    BNET_ENABLED = True
except Exception as ex:
    # We do get strange Google-defined exceptions from problems with protobuf, so
    # let's just catch (almost) everything. We do not want Lutris to crash, rather
    # we just want to suppress Battle.net and nothing else.
    logger.warning("The Battle.net source is unavailable because Google protobuf could not be loaded: %s", ex)
    BNET_ENABLED = False

GAME_IDS = {
    "s1": ("s1", "StarCraft", "S1", "starcraft-remastered"),
    "s2": ("s2", "StarCraft II", "S2", "starcraft-ii"),
    "wow": ("wow", "World of Warcraft", "WoW", "world-of-warcraft"),
    "wow_classic": ("wow_classic", "World of Warcraft Classic", "WoW_wow_classic", "world-of-warcraft-classic"),
    "wow_classic_era": (
        "wow_classic_era",
        "World of Warcraft Classic Era",
        "WoW_wow_classic_era",
        "world-of-warcraft-classic",
    ),
    "pro": ("pro", "Overwatch 2", "Pro", "overwatch-2"),
    "w1": ("w1", "Warcraft: Orcs & Humans", "W1", "warcraft-orcs-humans"),
    "w1r": ("w1r", "Warcraft I: Remastered", "W1R", "warcraft-i-remastered"),
    "w2bn": ("w2bn", "Warcraft II: Battle.net Edition", "W2BN", "warcraft-ii-battle-net-edition"),
    "w2r": ("w2r", "Warcraft II: Remastered", "W2R", "warcraft-ii-remastered"),
    "w3": ("w3", "Warcraft III", "W3", "warcraft-iii-reforged"),
    "gryphon": ("gryphon", "Warcraft Rumble", "GRY", "warcraft-rumble"),
    "hsb": ("hsb", "Hearthstone", "WTCG", "hearthstone"),
    "hero": ("hero", "Heroes of the Storm", "Hero", "heroes-of-the-storm"),
    "d3cn": ("d3cn", "暗黑破壞神III", "D3CN", "diablo-iii"),
    "d3": ("d3", "Diablo III", "D3", "diablo-iii"),
    "fenris": ("fenris", "Diablo IV", "Fen", "diablo-iv"),
    "viper": ("viper", "Call of Duty: Black Ops 4", "VIPR", "call-of-duty-black-ops-4"),
    "odin": ("odin", "Call of Duty: Modern Warfare", "ODIN", "call-of-duty-modern-warfare"),
    "lazarus": (
        "lazarus",
        "Call of Duty: MW2 Campaign Remastered",
        "LAZR",
        "call-of-duty-modern-warfare-2-campaign-remastered",
    ),
    "zeus": ("zeus", "Call of Duty: Black Ops Cold War", "ZEUS", "call-of-duty-black-ops-cold-war"),
    "auks": ("auks", "Call of Duty: Modern Warfare II", "AUKS", "call-of-duty-modern-warfare-ii"),
    "codhq": ("codhq", "Call of Duty HQ", "CODHQ", "call-of-duty-hq"),
    "rtro": ("rtro", "Blizzard Arcade Collection", "RTRO", "blizzard-arcade-collection"),
    "wlby": ("wlby", "Crash Bandicoot 4: It's About Time", "WLBY", "crash-bandicoot-4-its-about-time"),
    "osi": ("osi", "Diablo® II: Resurrected", "OSI", "diablo-2-ressurected"),
    "fore": ("fore", "Call of Duty: Vanguard", "FORE", "call-of-duty-vanguard"),
    "d2": ("d2", "Diablo® II", "Diablo II", "diablo-ii"),
    "d2LOD": ("d2LOD", "Diablo® II: Lord of Destruction®", "Diablo II", "diablo-ii-lord-of-destruction"),
    "w3ROC": ("w3ROC", "Warcraft® III: Reign of Chaos", "Warcraft III", "warcraft-iii-reign-of-chaos"),
    "w3tft": ("w3tft", "Warcraft® III: The Frozen Throne®", "Warcraft III", "warcraft-iii-the-frozen-throne"),
    "sca": ("sca", "StarCraft® Anthology", "Starcraft", "starcraft"),
    "anbs": ("anbs", "Diablo Immortal", "ANBS", "diablo-immortal"),
}


class BattleNetCover(ServiceMedia):
    service = "battlenet"
    size = (176, 234)
    file_patterns = ["%s.jpg"]
    dest_path = os.path.join(settings.CACHE_DIR, "battlenet/coverart")
    api_field = "coverart"


class BattleNetGame(ServiceGame):
    """Game from Battle.net"""

    service = "battlenet"
    runner = "wine"
    installer_slug = "battlenet"

    @classmethod
    def create(cls, blizzard_game):
        """Create a service game from an entry from the Dolphin cache"""
        service_game = cls()
        service_game.appid = blizzard_game[0]
        service_game.name = blizzard_game[1]
        service_game.slug = blizzard_game[3]
        service_game.details = json.dumps(
            {
                "id": blizzard_game[0],
                "name": blizzard_game[1],
                "product_code": blizzard_game[2],
                "slug": blizzard_game[3],
                "coverart": "https://lutris.net/games/cover/%s.jpg" % blizzard_game[3],
            }
        )
        return service_game


class BattleNetService(BaseService):
    """Service class for Battle.net"""

    id = "battlenet"
    name = _("Battle.net")
    icon = "battlenet"
    runner = "wine"
    medias = {"coverart": BattleNetCover}
    default_format = "coverart"
    client_installer = "battlenet"
    cookies_path = os.path.join(settings.CACHE_DIR, ".bnet.auth")
    cache_path = os.path.join(settings.CACHE_DIR, "bnet-library.json")
    redirect_uris = ["https://lutris.net"]

    @property
    def battlenet_config_path(self):
        return ""

    def load(self):
        games = [BattleNetGame.create(game) for game in GAME_IDS.values()]
        for game in games:
            game.save()
        return games

    def add_installed_games(self):
        """Scan an existing Battle.net install for games"""
        bnet_game = get_game_by_field(self.client_installer, "slug")
        if not bnet_game:
            raise RuntimeError("Battle.net is not installed in Lutris")
        bnet_prefix = bnet_game["directory"].split("drive_c")[0]
        parser = BlizzardProductDbParser(bnet_prefix)
        installed_slugs = []
        for game in parser.games:
            slug = self.install_from_battlenet(bnet_game, game)
            if slug:
                installed_slugs.append(slug)
        sync_media(installed_slugs)

    def install_from_battlenet(self, bnet_game, game):
        app_id = game.ngdp
        logger.debug("Installing Battle.net game %s", app_id)
        service_game = ServiceGameCollection.get_game("battlenet", app_id)
        if not service_game:
            logger.error("Aborting install, %s is not present in the game library.", app_id)
            return
        lutris_game_id = service_game["slug"] + "-" + self.id
        existing_game = get_game_by_field(lutris_game_id, "installer_slug")
        if existing_game:
            return
        game_config = LutrisConfig(game_config_id=bnet_game["configpath"]).game_level
        game_config["game"]["args"] = '--exec="launch %s"' % game.ngdp
        configpath = write_game_config(lutris_game_id, game_config)
        slug = service_game["slug"]
        add_game(
            name=service_game["name"],
            runner=bnet_game["runner"],
            slug=slug,
            directory=bnet_game["directory"],
            installed=1,
            installer_slug=lutris_game_id,
            configpath=configpath,
            service=self.id,
            service_id=app_id,
            platform="Windows",
        )
        return slug

    def get_installed_slug(self, db_game):
        return db_game["slug"]

    def generate_installer(self, db_game, bnet_db_game):
        bnet_app = Game(bnet_db_game["id"])
        bnet_exe = bnet_app.config.game_config["exe"]
        if not os.path.isabs(bnet_exe):
            bnet_exe = os.path.join(bnet_app.config.game_config["prefix"], bnet_exe)
        return {
            "name": db_game["name"],
            "version": self.name,
            "slug": db_game["slug"] + "-" + self.id,
            "game_slug": self.get_installed_slug(db_game),
            "runner": self.get_installed_runner_name(db_game),
            "appid": db_game["appid"],
            "script": {
                "requires": self.client_installer,
                "game": {
                    "args": '--exec="launch %s"' % db_game["appid"],
                },
                "installer": [
                    {
                        "task": {
                            "name": "wineexec",
                            "executable": bnet_exe,
                            "args": '--exec="install %s"' % db_game["appid"],
                            "prefix": bnet_app.config.game_config["prefix"],
                            "description": (
                                "Battle.net will now open. Please launch "
                                "the installation of %s then close Battle.net "
                                "once the game has been downloaded." % db_game["name"]
                            ),
                        }
                    }
                ],
            },
        }

    def get_installed_runner_name(self, db_game):
        return self.runner

    def install(self, db_game):
        bnet_game = get_game_by_field(self.client_installer, "slug")
        application = Gio.Application.get_default()
        application.show_installer_window(
            [self.generate_installer(db_game, bnet_game)], service=self, appid=db_game["appid"]
        )


class BlizzardProductDbParser:
    # Adapted from DatabaseParser in https://github.com/bartok765/galaxy_blizzard_plugin
    NOT_GAMES = ("bna", "agent")
    PRODUCT_DB_PATH = "/drive_c/ProgramData/Battle.net/Agent/product.db"

    def __init__(self, prefix_path):
        self.data = self.load_product_db(prefix_path + self.PRODUCT_DB_PATH)
        self.products = {}
        self._region = ""
        self.parse()

    @property
    def region(self):
        return self._region

    @staticmethod
    def load_product_db(product_db_path):
        with open(product_db_path, "rb") as f:
            pdb = f.read()
        return pdb

    @property
    def games(self):
        if self.products:
            return [v for k, v in self.products.items() if k not in self.NOT_GAMES]
        return []

    def parse(self):
        self.products = {}
        database = ProductDb()
        database.ParseFromString(self.data)

        for product_install in database.product_installs:  # pylint: disable=no-member
            # process region
            if product_install.product_code in ["agent", "bna"] and not self.region:
                self._region = product_install.settings.play_region

            ngdp_code = product_install.product_code
            uninstall_tag = product_install.uid
            install_path = product_install.settings.install_path
            playable = product_install.cached_product_state.base_product_state.playable
            version = product_install.cached_product_state.base_product_state.current_version_str
            installed = product_install.cached_product_state.base_product_state.installed

            self.products[ngdp_code] = ProductDbInfo(
                uninstall_tag, ngdp_code, install_path, version, playable, installed
            )



"""Shared config dialog stuff"""

# pylint: disable=not-an-iterable
from gettext import gettext as _

from gi.repository import Gtk

from lutris import settings
from lutris.config import LutrisConfig, make_game_config_id, rename_config
from lutris.game import Game
from lutris.gui.config import DIALOG_HEIGHT, DIALOG_WIDTH
from lutris.gui.config.boxes import GameBox, RunnerBox, SystemConfigBox
from lutris.gui.config.game_info_box import GameInfoBox
from lutris.gui.config.widget_generator import WidgetWarningMessageBox
from lutris.gui.dialogs import DirectoryDialog, ErrorDialog, QuestionDialog, SavableModelessDialog, display_error
from lutris.gui.dialogs.delegates import DialogInstallUIDelegate
from lutris.gui.dialogs.move_game import MoveDialog
from lutris.gui.widgets.notifications import send_notification
from lutris.runners import import_runner
from lutris.services.lutris import download_lutris_media
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger
from lutris.util.strings import parse_playtime, slugify


# pylint: disable=too-many-instance-attributes, no-member
class GameDialogCommon(SavableModelessDialog, DialogInstallUIDelegate):  # type:ignore[misc]
    """Base class for config dialogs"""

    no_runner_label = _("Select a runner in the Game Info tab")

    def __init__(self, title, config_level, parent=None):
        super().__init__(title, parent=parent, border_width=0)
        self.config_level = config_level
        self.set_default_size(DIALOG_WIDTH, DIALOG_HEIGHT)
        self.vbox.set_border_width(0)

        self.notebook = None

        self.info_box = None
        self.runner_box = None

        self.timer_id = None
        self.game = None
        self.saved = None
        self.option_page_indices = set()
        self.searchable_page_indices = set()
        self.advanced_switch_widgets = []
        self.header_bar_widgets = []
        self.game_box = None
        self.system_box = None
        self.runner_name = None
        self.lutris_config = None
        self.notebook_page_generators = {}
        self.notebook_page_updater = {}

        self.build_header_bar()

    @staticmethod
    def build_scrolled_window(widget):
        """Return a scrolled window containing config widgets"""
        scrolled_window = Gtk.ScrolledWindow(visible=True)
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.add(widget)
        return scrolled_window

    def build_notebook(self):
        self.notebook = Gtk.Notebook(visible=True)
        self.notebook.set_show_border(False)
        self.notebook.connect("switch-page", self.on_notebook_switch_page)
        self.vbox.pack_start(self.notebook, True, True, 0)

    def on_notebook_switch_page(self, notebook, page, index):
        generator = self.notebook_page_generators.get(index)
        if generator:
            generator()
            del self.notebook_page_generators[index]
        else:
            updater = self.notebook_page_updater.get(index)
            if updater:
                updater()

        self.update_advanced_switch_visibility(index)
        self.update_search_entry_visibility(index)

    def build_tabs(self):
        """Build tabs (for game and runner levels)"""
        self.timer_id = None
        if self.config_level == "game":
            self._build_info_tab()
            self._build_game_tab()
        self._build_runner_tab()
        self._build_system_tab()

        current_page_index = self.notebook.get_current_page()
        self.update_advanced_switch_visibility(current_page_index)
        self.update_search_entry_visibility(current_page_index)

    def set_header_bar_widgets_visibility(self, value):
        for widget in self.header_bar_widgets:
            widget.set_visible(value)

    def update_advanced_switch_visibility(self, current_page_index):
        if self.notebook:
            show_switch = current_page_index in self.option_page_indices
            for widget in self.advanced_switch_widgets:
                widget.set_visible(show_switch)

    def update_search_entry_visibility(self, current_page_index):
        """Shows or hides the search entry according to what page is currently displayed."""
        if self.notebook:
            show_search = current_page_index in self.searchable_page_indices
            self.set_search_entry_visibility(show_search)

    def set_search_entry_visibility(self, show_search, placeholder_text=None, tooltip_markup=None):
        """Explicitly shows or hides the search entry; can also update the placeholder text."""
        header_bar = self.get_header_bar()
        if show_search and self.search_entry:
            header_bar.set_custom_title(self.search_entry)
            self.search_entry.set_placeholder_text(placeholder_text or self.get_search_entry_placeholder())
            self.search_entry.set_tooltip_markup(tooltip_markup)
        else:
            header_bar.set_custom_title(None)

    def get_search_entry_placeholder(self):
        if self.game and self.game.name:
            return _("Search %s options") % self.game.name

        return _("Search options")

    def _build_info_tab(self):
        self.info_box = GameInfoBox(parent_widget=self, game=self.game)

        info_sw = self.build_scrolled_window(self.info_box)
        page_index = self._add_notebook_tab(info_sw, _("Game info"))
        self.option_page_indices.add(page_index)

    def on_move_clicked(self, _button):
        game_directory = self.game.directory if self.game else ""
        new_location = DirectoryDialog("Select new location for the game", default_path=game_directory, parent=self)
        if not new_location.folder or new_location.folder == game_directory:
            return
        move_dialog = MoveDialog(self.game, new_location.folder, parent=self)
        move_dialog.connect("game-moved", self.on_game_moved)
        move_dialog.move()

    def on_game_moved(self, dialog):
        """Show a notification when the game is moved"""
        new_directory = dialog.new_directory
        if new_directory:
            self.game = Game(self.game.id if self.game else None)
            self.lutris_config = self.game.config
            self._rebuild_tabs()
            if self.info_box.directory_entry:
                self.info_box.directory_entry.set_text(new_directory)
            send_notification("Finished moving game", "%s moved to %s" % (dialog.game, new_directory))
        else:
            send_notification("Failed to move game", "Lutris could not move %s" % dialog.game)

    def _build_game_tab(self):
        def is_searchable(game):
            return game.has_runner and len(game.runner.game_options) > 8

        def has_advanced(game):
            if game.has_runner:
                for opt in game.runner.game_options:
                    if opt.get("advanced"):
                        return True
            return False

        if self.game and self.runner_name:
            self.game.runner_name = self.runner_name
            self.game_box = self._build_options_tab(
                _("Game options"),
                lambda: GameBox(self.config_level, self.lutris_config, self.game),
                advanced=has_advanced(self.game),
                searchable=is_searchable(self.game),
            )
        elif self.runner_name:
            game = Game(None)
            game.runner_name = self.runner_name
            self.game_box = self._build_options_tab(
                _("Game options"),
                lambda: GameBox(self.config_level, self.lutris_config, game),
                advanced=has_advanced(game),
                searchable=is_searchable(game),
            )
        else:
            self._build_missing_options_tab(self.no_runner_label, _("Game options"))

    def _build_runner_tab(self):
        if self.runner_name:
            self.runner_box = self._build_options_tab(
                _("Runner options"), lambda: RunnerBox(self.config_level, self.lutris_config)
            )
        else:
            self._build_missing_options_tab(self.no_runner_label, _("Runner options"))

    def _build_system_tab(self):
        self.system_box = self._build_options_tab(
            _("System options"), lambda: SystemConfigBox(self.config_level, self.lutris_config)
        )

    def _build_options_tab(self, notebook_label, box_factory, advanced=True, searchable=True):
        if not self.lutris_config:
            raise RuntimeError("Lutris config not loaded yet")
        config_box = box_factory()
        page_index = self._add_notebook_tab(self.build_scrolled_window(config_box), notebook_label)

        self.notebook_page_updater[page_index] = config_box.update_widgets

        if page_index == 0:
            config_box.generate_widgets()
        else:
            self.notebook_page_generators[page_index] = config_box.generate_widgets

        if advanced:
            self.option_page_indices.add(page_index)
        if searchable:
            self.searchable_page_indices.add(page_index)
        return config_box

    def _build_missing_options_tab(self, missing_label, notebook_label):
        label = Gtk.Label(label=self.no_runner_label)
        page_index = self._add_notebook_tab(label, notebook_label)
        self.option_page_indices.add(page_index)

    def _add_notebook_tab(self, widget, label):
        return self.notebook.append_page(widget, Gtk.Label(label=label))

    def build_header_bar(self):
        self.search_entry = Gtk.SearchEntry(width_chars=30, placeholder_text=_("Search options"))
        self.search_entry.connect("search-changed", self.on_search_entry_changed)
        self.search_entry.show_all()

        # Advanced settings toggle
        switch_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5, no_show_all=True, visible=True)
        switch_box.set_tooltip_text(_("Show advanced options"))

        switch_label = Gtk.Label(_("Advanced"), no_show_all=True, visible=True)
        switch = Gtk.Switch(no_show_all=True, visible=True, valign=Gtk.Align.CENTER)
        switch.set_state(settings.read_setting("show_advanced_options") == "True")
        switch.connect("state-set", lambda _w, s: self.on_show_advanced_options_toggled(bool(s)))

        switch_box.pack_start(switch_label, False, False, 0)
        switch_box.pack_end(switch, False, False, 0)

        header_bar = self.get_header_bar()

        header_bar.pack_end(switch_box)

        # These lists need to be distinct, so they can be separately
        # hidden or shown without interfering with each other.
        self.advanced_switch_widgets = [switch_label, switch]
        self.header_bar_widgets = [self.cancel_button, self.save_button, switch_box]

        if self.notebook:
            self.update_advanced_switch_visibility(self.notebook.get_current_page())

    def on_search_entry_changed(self, entry):
        """Callback for the search input keypresses"""
        text = entry.get_text().lower().strip()
        self._set_filter(text)

    def on_show_advanced_options_toggled(self, is_active):
        settings.write_setting("show_advanced_options", is_active)

        self._set_advanced_options_visible(is_active)

    def _set_advanced_options_visible(self, value):
        """Change visibility of advanced options across all config tabs."""
        if self.info_box:
            self.info_box.advanced_visibility = value
        self.system_box.advanced_visibility = value
        if self.runner_box:
            self.runner_box.advanced_visibility = value
        if self.game_box:
            self.game_box.advanced_visibility = value

    def _set_filter(self, value):
        if self.system_box:
            self.system_box.filter = value
        if self.runner_box:
            self.runner_box.filter = value
        if self.game_box:
            self.game_box.filter = value

    def on_runner_changed(self, widget):
        """Action called when runner drop down is changed."""
        new_runner_index = widget.get_active()
        game_info_box = self.info_box
        if game_info_box.runner_index and new_runner_index != game_info_box.runner_index:
            dlg = QuestionDialog(
                {
                    "parent": self,
                    "question": _(
                        "Are you sure you want to change the runner for this game ? "
                        "This will reset the full configuration for this game and "
                        "is not reversible."
                    ),
                    "title": _("Confirm runner change"),
                }
            )

            if dlg.result == Gtk.ResponseType.YES:
                self._switch_runner(widget, new_runner_index)
            else:
                # Revert the dropdown menu to the previously selected runner
                widget.set_active(game_info_box.runner_index)
        else:
            self._switch_runner(widget, new_runner_index)

    def _switch_runner(self, widget, new_runner_index):
        """Rebuilds the UI on runner change"""
        current_page = self.notebook.get_current_page()
        game_info_box = self.info_box
        game_info_box.runner_index = new_runner_index
        if new_runner_index == 0:
            logger.info("No runner selected, resetting configuration")
            self.runner_name = None
            self.lutris_config = None
        else:
            runner_name = widget.get_model()[new_runner_index][1]
            if runner_name == self.runner_name:
                logger.debug("Runner unchanged, not creating a new config")
                return
            logger.info("Creating new configuration with runner %s", runner_name)
            self.runner_name = runner_name
            self.lutris_config = LutrisConfig(runner_slug=self.runner_name, level="game")
        self._rebuild_tabs()
        self.notebook.set_current_page(current_page)

    def _rebuild_tabs(self):
        """Rebuild notebook pages"""
        for i in range(self.notebook.get_n_pages(), 1, -1):
            self.notebook.remove_page(i - 1)
        self.option_page_indices.clear()
        self.searchable_page_indices.clear()
        self._build_game_tab()
        self._build_runner_tab()
        self._build_system_tab()
        self.show_all()

    def on_response(self, _widget, response):
        if response in (Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT):
            # Reload the config to clean out any changes we may have made
            if self.game:
                self.game.reload_config()
        super().on_response(_widget, response)

    def is_valid(self) -> bool:
        game_info_box = self.info_box
        if not self.runner_name:
            ErrorDialog(_("Runner not provided"), parent=self)
            return False
        if not game_info_box.name_entry.get_text():
            ErrorDialog(_("Please fill in the name"), parent=self)
            return False
        if self.runner_name == "steam" and not self.lutris_config.game_config.get("appid"):
            ErrorDialog(_("Steam AppID not provided"), parent=self)
            return False
        playtime_text = game_info_box.playtime_entry.get_text()
        if playtime_text and (not self.game or playtime_text != self.game.formatted_playtime):
            try:
                parse_playtime(playtime_text)
            except ValueError as ex:
                display_error(ex, parent=self)
                return False

        invalid_fields = []
        runner_class = import_runner(self.runner_name)
        runner_instance = runner_class()
        for config in ["game", "runner"]:
            for k, v in getattr(self.lutris_config, config + "_config").items():
                option = runner_instance.find_option(config + "_options", k)
                if option is None:
                    continue
                validator = option.get("validator")
                if validator is not None:
                    try:
                        res = validator(v)
                        logger.debug("%s validated successfully: %s", k, res)
                    except Exception:
                        invalid_fields.append(option.get("label"))
        if invalid_fields:
            ErrorDialog(_("The following fields have invalid values: ") + ", ".join(invalid_fields), parent=self)
            return False
        return True

    def on_save(self, _button):
        """Save game info and destroy widget."""
        if not self.is_valid():
            logger.warning(_("Current configuration is not valid, ignoring save request"))
            return
        game_info_box = self.info_box
        name = game_info_box.name_entry.get_text()
        sortname = game_info_box.sortname_entry.get_text()

        if not game_info_box.slug:
            game_info_box.slug = slugify(name)
        if game_info_box.slug != game_info_box.initial_slug:
            AsyncCall(download_lutris_media, None, game_info_box.slug)
        if not self.game:
            self.game = Game()

        year = None
        if game_info_box.year_entry.get_text():
            year = int(game_info_box.year_entry.get_text())

        playtime = None
        playtime_text = game_info_box.playtime_entry.get_text()
        if playtime_text and playtime_text != self.game.formatted_playtime:
            playtime = parse_playtime(playtime_text)

        if not self.lutris_config.game_config_id:
            self.lutris_config.game_config_id = make_game_config_id(game_info_box.slug)

        self.game.name = name
        self.game.sortname = sortname
        self.game.slug = game_info_box.slug
        self.game.year = year
        if playtime:
            self.game.playtime = playtime
        self.game.is_installed = True
        self.game.config = self.lutris_config

        # Rename config file if game slug changed
        if new_config_id := rename_config(self.game.config.game_config_id, self.game.slug):
            self.game.config.game_config_id = new_config_id

        self.game.runner_name = self.runner_name

        if "icon" not in self.game.custom_images:
            self.game.runner.extract_icon(game_info_box.slug)

        self.game.save()
        self.destroy()
        self.saved = True
        return True


class RunnerMessageBox(WidgetWarningMessageBox):
    def __init__(self):
        super().__init__(margin_left=12, margin_right=12, icon_name="dialog-warning")



from typing import Callable, Optional

from gi.repository import Gtk, Pango

from lutris.util.jobs import schedule_repeating_at_idle
from lutris.util.log import logger


class ProgressInfo:
    """Contains the current state of a process being monitored. This can also provide
    for stopping the process via a function you can provide.

    Processes sometimes cannot be stopped after a certain point; at that point they start
    providing Progress objects with no stop-function."""

    def __init__(self, progress: float = 0, label_markup: str = "", stop_function: Optional[Callable] = None):
        self.progress = progress
        self.label_markup = label_markup
        self.stop_function = stop_function
        self.has_ended = False

    @classmethod
    def ended(cls, label_markup: str = "") -> "ProgressInfo":
        """Creates a ProgressInfo whose has_ended flag is set, to indicate that
        the monitored process is over."""
        info = cls(1.0, label_markup=label_markup)
        info.has_ended = True
        return info

    @property
    def can_stop(self) -> bool:
        """Called to check if the stop button should appear."""
        return bool(self.stop_function)

    def stop(self):
        """Called whe the stop button is clicked."""
        if self.stop_function:
            try:
                self.stop_function()
            except Exception as ex:
                logger.exception("Error during progress box stop: %s", ex)


class ProgressBox(Gtk.Box):
    """Simple, small progress bar used to monitor the update of runtime or runner components.
    This class needs only a function that returns a Progress object, which describes the current
    progress and optionally can stop the update."""

    ProgressFunction = Callable[[], "ProgressInfo"]

    def __init__(self, progress_function: ProgressFunction, **kwargs):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, no_show_all=True, spacing=6, **kwargs)

        self.progress_function = progress_function
        self.progress = ProgressInfo(0.0)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True, spacing=6, valign=Gtk.Align.CENTER)

        self.label = Gtk.Label(label="", visible=False, wrap=True, ellipsize=Pango.EllipsizeMode.MIDDLE, xalign=0)
        vbox.pack_start(self.label, False, False, 0)

        self.progressbar = Gtk.ProgressBar(pulse_step=0.4, visible=True)
        self.progressbar.set_valign(Gtk.Align.CENTER)
        vbox.pack_start(self.progressbar, False, False, 0)

        self.pack_start(vbox, True, True, 0)

        self.stop_button = Gtk.Button.new_from_icon_name("media-playback-stop-symbolic", Gtk.IconSize.BUTTON)
        self.stop_button.hide()
        self.stop_button.get_style_context().add_class("circular")
        self.stop_button.connect("clicked", self.on_stop_clicked)
        self.pack_start(self.stop_button, False, False, 0)

        self._destroyed = False
        self._apply_progress(ProgressInfo(0.0, "Please wait..."))
        self._timer_task = schedule_repeating_at_idle(self.on_update_progress, interval_seconds=0.5)
        self.connect("destroy", self.on_destroy)

    def on_stop_clicked(self, _widget) -> None:
        if self.progress.can_stop:
            self.progress.stop()

    def on_destroy(self, _widget) -> None:
        self._destroyed = True
        self._timer_task.unschedule()

    def on_update_progress(self) -> bool:
        try:
            self.update_progress()
            return True
        except Exception as ex:
            logger.exception("Unable to obtain a progress update: %s", ex)
            return False

    def update_progress(self) -> None:
        """Invokes the progress function and displays what it returns;
        this can be called to ensure the box is immediately up-to-date,
        without waiting for idle-time."""
        progress = self.progress_function()
        self._apply_progress(progress)

    def _apply_progress(self, progress: ProgressInfo):
        # Just in case the progress-function destroys the progress box.
        if self._destroyed:
            return

        self.progress = progress

        if progress.progress is None:
            self.progressbar.pulse()
        else:
            self.progressbar.set_fraction(min(progress.progress, 1))
        self._set_label(progress.label_markup or "")
        self.stop_button.set_visible(progress.can_stop)

    def _set_label(self, markup: str) -> None:
        if markup:
            if markup != self.label.get_text():
                self.label.set_markup(markup)
            self.label.show()
        else:
            self.label.hide()



import os
import shutil
from gettext import gettext as _

from lutris import settings
from lutris.runners.runner import Runner


class redream(Runner):
    human_name = _("Redream")
    description = _("Sega Dreamcast emulator")
    platforms = [_("Sega Dreamcast")]
    runner_executable = "redream/redream"
    download_url = "https://redream.io/download/redream.x86_64-linux-v1.5.0.tar.gz"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("Disc image file"),
            "help": _("Game data file\nSupported formats: GDI, CDI, CHD"),
        }
    ]
    runner_options = [
        {"option": "fs", "type": "bool", "section": _("Graphics"), "label": _("Fullscreen"), "default": False},
        {
            "option": "ar",
            "type": "choice",
            "section": _("Graphics"),
            "label": _("Aspect Ratio"),
            "choices": [(_("4:3"), "4:3"), (_("Stretch"), "stretch")],
            "default": "4:3",
        },
        {
            "option": "region",
            "type": "choice",
            "label": _("Region"),
            "choices": [(_("USA"), "usa"), (_("Europe"), "europe"), (_("Japan"), "japan")],
            "default": "usa",
        },
        {
            "option": "language",
            "type": "choice",
            "label": _("System Language"),
            "choices": [
                (_("English"), "english"),
                (_("German"), "german"),
                (_("French"), "french"),
                (_("Spanish"), "spanish"),
                (_("Italian"), "italian"),
                (_("Japanese"), "japanese"),
            ],
            "default": "english",
        },
        {
            "option": "broadcast",
            "type": "choice",
            "label": "Television System",
            "choices": [
                (_("NTSC"), "ntsc"),
                (_("PAL"), "pal"),
                (_("PAL-M (Brazil)"), "pal_m"),
                (_("PAL-N (Argentina, Paraguay, Uruguay)"), "pal_n"),
            ],
            "default": "ntsc",
        },
        {
            "option": "time_sync",
            "type": "choice",
            "label": _("Time Sync"),
            "choices": [
                (_("Audio and video"), "audio and video"),
                (_("Audio"), "audio"),
                (_("Video"), "video"),
                (_("None"), "none"),
            ],
            "default": "audio and video",
            "advanced": True,
        },
        {
            "option": "int_res",
            "type": "choice",
            "label": _("Internal Video Resolution Scale"),
            "choices": [
                ("×1", "1"),
                ("×2", "2"),
                ("×3", "3"),
                ("×4", "4"),
                ("×5", "5"),
                ("×6", "6"),
                ("×7", "7"),
                ("×8", "8"),
            ],
            "default": "2",
            "advanced": True,
            "help": _("Only available in premium version."),
        },
    ]

    def install(self, install_ui_delegate, version=None, callback=None):
        def on_runner_installed(*args):
            license_filename = install_ui_delegate.show_install_file_inquiry(
                question=_("Do you want to select a premium license file?"),
                title=_("Use premium version?"),
                message=_("Use premium version?"),
            )

            if license_filename:
                shutil.copy(license_filename, os.path.join(settings.RUNNER_DIR, "redream"))

        super().install(install_ui_delegate, version=version, callback=on_runner_installed)

    def play(self):
        command = self.get_command()

        if self.runner_config.get("fs") is True:
            command.append("--fullscreen=1")
        else:
            command.append("--fullscreen=0")

        if self.runner_config.get("ar"):
            command.append("--aspect=" + self.runner_config.get("ar"))

        if self.runner_config.get("region"):
            command.append("--region=" + self.runner_config.get("region"))

        if self.runner_config.get("language"):
            command.append("--language=" + self.runner_config.get("language"))

        if self.runner_config.get("broadcast"):
            command.append("--broadcast=" + self.runner_config.get("broadcast"))

        if self.runner_config.get("time_sync"):
            command.append("--time_sync=" + self.runner_config.get("time_sync"))

        if self.runner_config.get("int_res"):
            command.append("--res=" + self.runner_config.get("int_res"))

        command.append(self.game_config.get("main_file"))

        return {"command": command}



"""Manage Humble Bundle libraries"""

import concurrent.futures
import json
import os
from gettext import gettext as _

from gi.repository import Gtk

from lutris import settings
from lutris.exceptions import UnavailableGameError
from lutris.gui.dialogs import HumbleBundleCookiesDialog, QuestionDialog
from lutris.installer import AUTO_ELF_EXE, AUTO_WIN32_EXE
from lutris.installer.installer_file import InstallerFile
from lutris.services.base import SERVICE_LOGIN, OnlineService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import linux
from lutris.util.http import HTTPError, Request
from lutris.util.log import logger


class HumbleBundleIcon(ServiceMedia):
    """HumbleBundle icon"""

    service = "humblebundle"
    size = (70, 70)
    dest_path = os.path.join(settings.CACHE_DIR, "humblebundle/icons")
    file_patterns = ["%s.png"]
    api_field = "icon"


class HumbleSmallIcon(HumbleBundleIcon):
    size = (35, 35)
    can_be_fallback = False

    def get_fallback_media_path(self, services):
        return None


class HumbleBigIcon(HumbleBundleIcon):
    size = (105, 105)


class HumbleBundleGame(ServiceGame):
    """Service game for DRM free Humble Bundle games"""

    service = "humblebundle"

    @classmethod
    def new_from_humble_game(cls, humble_game):
        """Converts a game from the API to a service game usable by Lutris"""
        service_game = HumbleBundleGame()
        service_game.appid = humble_game["machine_name"]
        service_game.slug = humble_game["machine_name"]
        service_game.name = humble_game["human_name"]
        service_game.details = json.dumps(humble_game)
        return service_game


class HumbleBundleService(OnlineService):
    """Service for Humble Bundle"""

    id = "humblebundle"
    _matcher = "humble"
    name = _("Humble Bundle")
    icon = "humblebundle"
    online = True
    drm_free = True
    medias = {"small_icon": HumbleSmallIcon, "icon": HumbleBundleIcon, "big_icon": HumbleBigIcon}
    default_format = "icon"

    api_url = "https://www.humblebundle.com/"
    login_url = "https://www.humblebundle.com/login?goto=/home/library"
    redirect_uris = ["https://www.humblebundle.com/home/library"]

    cookies_path = os.path.join(settings.CACHE_DIR, ".humblebundle.auth")
    token_path = os.path.join(settings.CACHE_DIR, ".humblebundle.token")
    cache_path = os.path.join(settings.CACHE_DIR, "humblebundle/library/")

    supported_platforms = ("linux", "windows")

    def login(self, parent=None):
        dialog = QuestionDialog(
            {
                "title": _("Workaround for Humble Bundle authentication"),
                "question": _(
                    "Humble Bundle is restricting API calls from software like Lutris and GameHub.\n"
                    "Authentication to the service will likely fail.\n"
                    "There is a workaround involving copying cookies "
                    "from Firefox, do you want to do this instead?"
                ),
                "parent": parent,
            }
        )
        if dialog.result == Gtk.ResponseType.YES:
            dialog = HumbleBundleCookiesDialog()
            if dialog.cookies_content:
                with open(self.cookies_path, "w", encoding="utf-8") as cookies_file:
                    cookies_file.write(dialog.cookies_content)
                SERVICE_LOGIN.fire(self)
            else:
                self.logout()
        else:
            return super().login(parent=parent)

    def login_callback(self, url):
        """Called after the user has logged in successfully"""
        SERVICE_LOGIN.fire(self)

    def is_connected(self):
        """This doesn't actually check if the authentication
        is valid like the GOG service does.
        """
        return self.is_authenticated()

    def load(self):
        """Load the user's Humble Bundle library"""
        try:
            library = self.get_library()
        except ValueError as ex:
            raise RuntimeError("Failed to get Humble Bundle library. Try logging out and back-in.") from ex
        humble_games = []
        seen = set()
        for game in library:
            if game["human_name"] in seen:
                continue
            humble_games.append(HumbleBundleGame.new_from_humble_game(game))
            seen.add(game["human_name"])
        for game in humble_games:
            game.save()
        return humble_games

    def make_api_request(self, url):
        """Make an authenticated request to the Humble API"""
        request = Request(url, cookies=self.load_cookies())
        try:
            request.get()
        except HTTPError:
            logger.error("Failed to request %s", url)
            return
        return request.json

    def order_path(self, gamekey):
        """Return the local path for an order"""
        return os.path.join(self.cache_path, "%s.json" % gamekey)

    def get_order(self, gamekey):
        """Retrieve an order identitied by its key"""
        # logger.debug("Getting Humble Bundle order %s", gamekey)
        cache_filename = self.order_path(gamekey)
        if os.path.exists(cache_filename):
            with open(cache_filename, encoding="utf-8") as cache_file:
                return json.load(cache_file)
        response = self.make_api_request(self.api_url + "api/v1/order/%s?all_tpkds=true" % gamekey)
        os.makedirs(self.cache_path, exist_ok=True)
        with open(cache_filename, "w", encoding="utf-8") as cache_file:
            json.dump(response, cache_file)
        return response

    def get_library(self):
        """Return the games from the user's library"""
        games = []
        for order in self.get_orders():
            if not order:
                continue
            for product in order["subproducts"]:
                for download in product["downloads"]:
                    if download["platform"] in self.supported_platforms:
                        games.append(product)
        return games

    def get_gamekeys_from_local_orders(self):
        """Retrieve a list of orders from the cache."""
        game_keys = []
        if os.path.exists(self.cache_path):
            for order_file in os.listdir(self.cache_path):
                if not order_file.endswith(".json"):
                    continue
                game_keys.append({"gamekey": order_file[:-5]})
        return game_keys

    def get_orders(self):
        """Return all orders"""
        gamekeys = self.get_gamekeys_from_local_orders()
        orders = []
        if not gamekeys:
            gamekeys = self.make_api_request(self.api_url + "api/v1/user/order")
        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
            future_orders = [executor.submit(self.get_order, gamekey["gamekey"]) for gamekey in gamekeys]
            for order in future_orders:
                orders.append(order.result())
        logger.info("Loaded %s Humble Bundle orders", len(orders))
        return orders

    @staticmethod
    def find_download_in_order(order, humbleid, platform):
        """Return the download information in an order for a give game"""
        for product in order["subproducts"]:
            if product["machine_name"] != humbleid:
                continue
            available_platforms = [d["platform"] for d in product["downloads"]]
            if platform not in available_platforms:
                logger.warning(
                    "Requested platform %s not available in available platforms: %s", platform, available_platforms
                )

                if "linux" in available_platforms:
                    platform = "linux"
                elif "windows" in available_platforms:
                    platform = "windows"
                else:
                    platform = available_platforms[0]
            for download in product["downloads"]:
                if download["platform"] != platform:
                    continue
                return {
                    "product": order["product"],
                    "gamekey": order["gamekey"],
                    "created": order["created"],
                    "download": download,
                }

    def get_downloads(self, humbleid, platform):
        """Return the download information for a given game"""
        download_links = []
        for order in self.get_orders():
            download = self.find_download_in_order(order, humbleid, platform)
            if download:
                download_links.append(download)
        return download_links

    def get_installer_files(self, installer, installer_file_id, _selected_extras):
        """Replace the user provided file with download links from Humble Bundle"""
        try:
            link = get_humble_download_link(installer.service_appid, installer.runner)
        except Exception as ex:
            logger.exception("Failed to get Humble Bundle game: %s", ex)
            raise UnavailableGameError(_("The download URL for the game could not be determined.")) from ex
        if not link:
            raise UnavailableGameError(_("No game found on Humble Bundle"))
        filename = link.split("?")[0].split("/")[-1]
        file = InstallerFile(installer.game_slug, installer_file_id, {"url": link, "filename": filename})
        return [file], []

    @staticmethod
    def get_filename_for_platform(downloads, platform):
        download = [d for d in downloads if d["platform"] == platform][0]
        url = pick_download_url_from_download_info(download)
        if not url:
            return
        return url.split("?")[0].split("/")[-1]

    @staticmethod
    def platform_has_downloads(downloads, platform):
        for download in downloads:
            if download["platform"] != platform:
                continue
            if len(download["download_struct"]) > 0:
                return True

    def generate_installer(self, db_game):
        details = json.loads(db_game["details"])
        platforms = [download["platform"] for download in details["downloads"]]
        system_config = {}
        if "linux" in platforms and self.platform_has_downloads(details["downloads"], "linux"):
            runner = "linux"
            game_config = {"exe": AUTO_ELF_EXE}
            filename = self.get_filename_for_platform(details["downloads"], "linux")
            if filename.lower().endswith(".sh"):
                script = [
                    {"extract": {"file": "humblegame", "format": "zip", "dst": "$CACHE"}},
                    {"merge": {"src": "$CACHE/data/noarch", "dst": "$GAMEDIR", "optional": True}},
                    {"move": {"src": "$CACHE/data/noarch", "dst": "$CACHE/noarch", "optional": True}},
                    {"merge": {"src": "$CACHE/data/x86_64", "dst": "$GAMEDIR", "optional": True}},
                    {"move": {"src": "$CACHE/data/x86_64", "dst": "$CACHE/x86_64", "optional": True}},
                    {"merge": {"src": "$CACHE/data/x86", "dst": "$GAMEDIR", "optional": True}},
                    {"move": {"src": "$CACHE/data/x86", "dst": "$CACHE/x86", "optional": True}},
                    {"merge": {"src": "$CACHE/data/", "dst": "$GAMEDIR", "optional": True}},
                ]
            elif filename.endswith("-bin") or filename.endswith("mojo.run"):
                script = [
                    {"extract": {"file": "humblegame", "format": "zip", "dst": "$CACHE"}},
                    {"merge": {"src": "$CACHE/data/", "dst": "$GAMEDIR"}},
                ]
            elif filename.endswith(".air"):
                script = [
                    {"move": {"src": "humblegame", "dst": "$GAMEDIR"}},
                ]
            else:
                script = [{"extract": {"file": "humblegame"}}]
                system_config = {"gamemode": "false"}  # Unity games crash with gamemode
        elif "windows" in platforms:
            runner = "wine"
            game_config = {"exe": AUTO_WIN32_EXE, "prefix": "$GAMEDIR"}
            filename = self.get_filename_for_platform(details["downloads"], "windows")
            if filename.lower().endswith(".zip"):
                script = [
                    {"task": {"name": "create_prefix", "prefix": "$GAMEDIR"}},
                    {"extract": {"file": "humblegame", "dst": "$GAMEDIR/drive_c/%s" % db_game["slug"]}},
                ]
            else:
                script = [{"task": {"name": "wineexec", "executable": "humblegame"}}]
        else:
            logger.warning("Unsupported platforms: %s", platforms)
            return {}
        return {
            "name": db_game["name"],
            "version": "Humble Bundle",
            "slug": details["machine_name"],
            "game_slug": self.get_installed_slug(db_game),
            "runner": runner,
            "humbleid": db_game["appid"],
            "script": {
                "game": game_config,
                "system": system_config,
                "files": [{"humblegame": "N/A:Select the installer from Humble Bundle"}],
                "installer": script,
            },
        }

    def get_installed_runner_name(self, db_game):
        details = json.loads(db_game["details"])
        platforms = [download["platform"] for download in details["downloads"]]

        if "linux" in platforms and self.platform_has_downloads(details["downloads"], "linux"):
            return "linux"

        if "windows" in platforms:
            return "wine"

        return ""


def pick_download_url_from_download_info(download_info):
    """From a list of downloads in Humble Bundle, pick the most appropriate one
    for the installer.
    This needs a way to be explicitely filtered.
    """
    if not download_info["download_struct"]:
        logger.warning("No downloads found")
        return

    def humble_sort(download):
        name = download["name"]
        if "rpm" in name:
            return -99  # Not supported as an extractor
        bonus = 1
        if "deb" not in name:
            bonus = 2
        if linux.LINUX_SYSTEM.is_64_bit:
            if "386" in name or "32" in name:
                return -1
        else:
            if "64" in name:
                return -10
        return 1 * bonus

    sorted_downloads = sorted(download_info["download_struct"], key=humble_sort, reverse=True)
    logger.debug("Humble bundle installers:")
    for download in sorted_downloads:
        logger.debug(download)
    return sorted_downloads[0]["url"]["web"]


def get_humble_download_link(humbleid, runner):
    """Return a download link for a given humbleid and runner"""
    service = HumbleBundleService()
    platform = runner if runner != "wine" else "windows"
    downloads = service.get_downloads(humbleid, platform)
    if not downloads:
        logger.error("Game %s for %s not found in the Humble Bundle library", humbleid, platform)
        return
    logger.info("Found %s download for %s", len(downloads), humbleid)
    download = downloads[0]
    logger.info("Reloading order %s", download["product"]["human_name"])
    os.remove(service.order_path(download["gamekey"]))
    order = service.get_order(download["gamekey"])
    download_info = service.find_download_in_order(order, humbleid, platform)
    if download_info:
        return pick_download_url_from_download_info(download_info["download"])
    logger.warning("Couldn't retrieve any downloads for %s", humbleid)



"""Migrate banners and coverart from .cache/lutris to .local/share/lutris"""

import os

from lutris import settings
from lutris.util.log import logger


def _migrate(dirname):
    dest_dir = os.path.join(settings.DATA_DIR, dirname)
    src_dir = os.path.join(settings.CACHE_DIR, dirname)

    try:
        # init_lutris() creates the new banners directory
        if os.path.isdir(src_dir) and os.path.isdir(dest_dir):
            for filename in os.listdir(src_dir):
                src_file = os.path.join(src_dir, filename)
                dest_file = os.path.join(dest_dir, filename)

                if not os.path.exists(dest_file):
                    os.rename(src_file, dest_file)
                else:
                    os.unlink(src_file)

            if not os.listdir(src_dir):
                os.rmdir(src_dir)
    except OSError as ex:
        logger.exception("Failed to migrate banners: %s", ex)


def migrate():
    _migrate("banners")
    _migrate("coverart")



from gettext import gettext as _

from gi.repository import Gtk

from lutris.cache import get_custom_cache_path, save_custom_cache_path
from lutris.gui.dialogs import ModalDialog
from lutris.gui.widgets.common import FileChooserEntry


class CacheConfigurationDialog(ModalDialog):
    def __init__(self, parent=None):
        super().__init__(_("Download cache configuration"), parent=parent, flags=Gtk.DialogFlags.MODAL, border_width=10)
        self.timer_id = None
        self.set_size_request(480, 150)

        self.cache_path = get_custom_cache_path() or ""
        self.get_content_area().add(self.get_cache_config())

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_default_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)

        self.show_all()
        result = self.run()

        if result == Gtk.ResponseType.OK:
            save_custom_cache_path(self.cache_path)

        self.destroy()

    def get_cache_config(self):
        """Return the widgets for the cache configuration"""
        prefs_box = Gtk.VBox()

        box = Gtk.Box(spacing=12, margin_right=12, margin_left=12)
        label = Gtk.Label(_("Cache path"))
        box.pack_start(label, False, False, 0)
        path_chooser = FileChooserEntry(
            title=_("Set the folder for the cache path"),
            action=Gtk.FileChooserAction.SELECT_FOLDER,
            warn_if_non_writable_parent=True,
            text=self.cache_path,
            activates_default=True,
        )
        path_chooser.connect("changed", self._on_cache_path_set)
        box.pack_start(path_chooser, True, True, 0)

        prefs_box.pack_start(box, False, False, 6)
        cache_help_label = Gtk.Label(visible=True)
        cache_help_label.set_size_request(400, -1)
        cache_help_label.set_markup(
            _(
                "If provided, this location will be used by installers to cache "
                "downloaded files locally for future re-use. \nIf left empty, the "
                "installer files are discarded after the install completion."
            )
        )
        prefs_box.pack_start(cache_help_label, False, False, 6)
        return prefs_box

    def _on_cache_path_set(self, entry):
        self.cache_path = entry.get_text()



"""Module to deal with various aspects of displays"""

# isort:skip_file
import enum
import os
import subprocess
from typing import Any, Dict, Optional

import gi

try:
    gi.require_version("GnomeDesktop", "3.0")
    from gi.repository import GnomeDesktop  # type: ignore

    LIB_GNOME_DESKTOP_AVAILABLE = True
except ValueError:
    LIB_GNOME_DESKTOP_AVAILABLE = False
    GnomeDesktop = None

try:
    from dbus.exceptions import DBusException

    DBUS_AVAILABLE = True
except ImportError:
    DBUS_AVAILABLE = False

from gi.repository import Gdk, GLib, Gio, Gtk

from lutris.util import cache_single
from lutris.settings import DEFAULT_RESOLUTION_HEIGHT, DEFAULT_RESOLUTION_WIDTH
from lutris.util.graphics.displayconfig import MutterDisplayManager
from lutris.util.graphics.xrandr import LegacyDisplayManager, change_resolution, get_outputs
from lutris.util.log import logger


def get_default_dpi():
    """Computes the DPI to use for the primary monitor
    which we pass to WINE."""
    display = Gdk.Display.get_default()
    if display:
        monitor = display.get_primary_monitor()
        if monitor:
            scale = monitor.get_scale_factor()
            dpi = 96 * scale
            return int(dpi)
    return 96


@cache_single
def is_display_x11():
    """True if"""
    display = Gdk.Display.get_default()
    return "x11" in type(display).__name__.casefold()


class DisplayManager:
    """Get display and resolution using GnomeDesktop"""

    def __init__(self, screen: Gdk.Screen):
        self.rr_screen = GnomeDesktop.RRScreen.new(screen)
        self.rr_config = GnomeDesktop.RRConfig.new_current(self.rr_screen)
        self.rr_config.load_current()

    def get_display_names(self):
        """Return names of connected displays"""
        return [output_info.get_display_name() for output_info in self.rr_config.get_outputs()]

    def get_resolutions(self):
        """Return available resolutions"""
        resolutions = ["%sx%s" % (mode.get_width(), mode.get_height()) for mode in self.rr_screen.list_modes()]
        if not resolutions:
            logger.error("Failed to generate resolution list from default GdkScreen")
            return ["%sx%s" % (DEFAULT_RESOLUTION_WIDTH, DEFAULT_RESOLUTION_HEIGHT)]
        return sorted(set(resolutions), key=lambda x: int(x.split("x")[0]), reverse=True)

    def _get_primary_output(self):
        """Return the RROutput used as a primary display"""
        for output in self.rr_screen.list_outputs():
            if output.get_is_primary():
                return output
        return

    def get_current_resolution(self):
        """Return the current resolution for the primary display"""
        output = self._get_primary_output()
        if not output:
            logger.error("Failed to get a default output")
            return str(DEFAULT_RESOLUTION_WIDTH), str(DEFAULT_RESOLUTION_HEIGHT)
        current_mode = output.get_current_mode()
        return str(current_mode.get_width()), str(current_mode.get_height())

    @staticmethod
    def set_resolution(resolution):
        """Set the resolution of one or more displays.
        The resolution can either be a string, which will be applied to the
        primary display or a list of configurations as returned by `get_config`.
        This method uses XrandR and will not work on Wayland.
        """
        return change_resolution(resolution)

    @staticmethod
    def get_config():
        """Return the current display resolution
        This method uses XrandR and will not work on wayland
        The output can be fed in `set_resolution`
        """
        return get_outputs()


def get_display_manager():
    """Return the appropriate display manager instance.
    Defaults to Mutter if available. This is the only one to support Wayland.
    """
    if DBUS_AVAILABLE:
        try:
            return MutterDisplayManager()
        except DBusException as ex:
            logger.debug("Mutter DBus service not reachable: %s", ex)
        except Exception as ex:  # pylint: disable=broad-except
            logger.exception("Failed to instantiate MutterDisplayConfig. Please report with exception: %s", ex)
    else:
        logger.error("DBus is not available, Lutris was not properly installed.")

    if LIB_GNOME_DESKTOP_AVAILABLE:
        try:
            screen = Gdk.Screen.get_default()
            if screen:
                return DisplayManager(screen)
        except GLib.Error:
            pass
    return LegacyDisplayManager()


DISPLAY_MANAGER = get_display_manager()


class DesktopEnvironment(enum.Enum):
    """Enum of desktop environments."""

    PLASMA = 0
    MATE = 1
    XFCE = 2
    DEEPIN = 3
    UNKNOWN = 999


# These desktop environment use a compositor that can be detected with a specific
# command, and which provide a definite answer; the DE can be asked to start and stop it..
_compositor_commands_by_de = {
    DesktopEnvironment.MATE: {
        "check": ["gsettings", "get", "org.mate.Marco.general", "compositing-manager"],
        "active_result": b"true\n",
        "stop_compositor": ["gsettings", "set", "org.mate.Marco.general", "compositing-manager", "false"],
        "start_compositor": ["gsettings", "set", "org.mate.Marco.general", "compositing-manager", "true"],
    },
    DesktopEnvironment.XFCE: {
        "check": ["xfconf-query", "--channel=xfwm4", "--property=/general/use_compositing"],
        "active_result": b"true\n",
        "stop_compositor": ["xfconf-query", "--channel=xfwm4", "--property=/general/use_compositing", "--set=false"],
        "start_compositor": ["xfconf-query", "--channel=xfwm4", "--property=/general/use_compositing", "--set=true"],
    },
    DesktopEnvironment.DEEPIN: {
        "check": [
            "dbus-send",
            "--session",
            "--dest=com.deepin.WMSwitcher",
            "--type=method_call",
            "--print-reply=literal",
            "/com/deepin/WMSwitcher",
            "com.deepin.WMSwitcher.CurrentWM",
        ],
        "active_result": b"deepin wm\n",
        "stop_compositor": [
            "dbus-send",
            "--session",
            "--dest=com.deepin.WMSwitcher",
            "--type=method_call",
            "/com/deepin/WMSwitcher",
            "com.deepin.WMSwitcher.RequestSwitchWM",
        ],
        "start_compositor": [
            "dbus-send",
            "--session",
            "--dest=com.deepin.WMSwitcher",
            "--type=method_call",
            "/com/deepin/WMSwitcher",
            "com.deepin.WMSwitcher.RequestSwitchWM",
        ],
    },
}

# These additional compositors can be detected by looking for their process,
# and must be started more directly.
_non_de_compositor_commands = [
    {
        "check": ["pgrep", "picom"],
        "stop_compositor": ["pkill", "picom"],
        "start_compositor": ["picom", ""],
        "run_in_background": True,
    },
    {
        "check": ["pgrep", "compton"],
        "stop_compositor": ["pkill", "compton"],
        "start_compositor": ["compton", ""],
        "run_in_background": True,
    },
]


def get_desktop_environment() -> Optional[DesktopEnvironment]:
    """Converts the value of the DESKTOP_SESSION environment variable
    to one of the constants in the DesktopEnvironment class.
    Returns None if DESKTOP_SESSION is empty or unset.
    """
    desktop_session = os.environ.get("DESKTOP_SESSION", "").lower()
    if not desktop_session:
        return None
    if desktop_session.endswith("mate"):
        return DesktopEnvironment.MATE
    if desktop_session.endswith("xfce"):
        return DesktopEnvironment.XFCE
    if desktop_session.endswith("deepin"):
        return DesktopEnvironment.DEEPIN
    if "plasma" in desktop_session:
        return DesktopEnvironment.PLASMA
    return DesktopEnvironment.UNKNOWN


def _get_command_output(command):
    """Some rogue function that gives no shit about residing in the correct module"""
    try:
        return subprocess.Popen(  # pylint: disable=consider-using-with
            command, stdin=subprocess.DEVNULL, stdout=subprocess.PIPE, close_fds=True
        ).communicate()[0]
    except FileNotFoundError:
        logger.error("Unable to run command, %s not found", command[0])


def is_compositing_enabled() -> bool:
    """Checks whether compositing is currently disabled or enabled.
    Returns True for enabled, False for disabled or if we didn't recognize
    the compositor.
    """

    desktop_environment = get_desktop_environment()
    if desktop_environment in _compositor_commands_by_de:
        command_set = _compositor_commands_by_de[desktop_environment]
        return _check_compositor_active(command_set)

    for command_set in _non_de_compositor_commands:
        if _check_compositor_active(command_set):
            return True

    # No compositor detected
    return False


def _check_compositor_active(command_set: Dict[str, Any]) -> bool:
    """Applies the 'check' command; and returns whether the result
    was the desired 'active_result'; if that is omitted, we check for
    any result at all."""
    command = command_set["check"]
    result = _get_command_output(command)

    if "active_result" in command_set:
        return result == command_set["active_result"]

    return result != b""


# One element is appended to this for every invocation of disable_compositing:
# True if compositing has been disabled, False if not. enable_compositing
# removes the last element, and only re-enables compositing if that element
# was True.
_COMPOSITING_DISABLED_STACK = []


@cache_single
def _get_compositor_commands():
    """Returns the commands to enable/disable compositing on the current
    desktop environment as a 3-tuple: start command, stop-command and
    a flag to indicate if we need to run the commands in the background.
    """
    desktop_environment = get_desktop_environment()
    command_set = None
    if desktop_environment is not None:
        command_set = _compositor_commands_by_de.get(desktop_environment)

    if not command_set:
        for c in _non_de_compositor_commands:
            if _check_compositor_active(c):
                command_set = c
                break

    if command_set:
        start_compositor = command_set["start_compositor"]
        stop_compositor = command_set["stop_compositor"]
        run_in_background = bool(command_set.get("run_in_background"))
        return start_compositor, stop_compositor, run_in_background

    return None, None, False


def _run_command(*command, run_in_background=False):
    """Random _run_command lost in the middle of the project,
    are you lost little _run_command?
    """
    try:
        if run_in_background:
            command = " ".join(command)
        return subprocess.Popen(  # pylint: disable=consider-using-with
            command,
            stdin=subprocess.DEVNULL,
            close_fds=True,
            shell=run_in_background,
            start_new_session=run_in_background,
        )
    except FileNotFoundError:
        errorMessage = "FileNotFoundError when running command:", command
        logger.error(errorMessage)


def disable_compositing():
    """Disable compositing if not already disabled."""
    compositing_enabled = is_compositing_enabled()
    if any(_COMPOSITING_DISABLED_STACK):
        compositing_enabled = False
    _COMPOSITING_DISABLED_STACK.append(compositing_enabled)
    if not compositing_enabled:
        return
    _, stop_compositor, background = _get_compositor_commands()
    if stop_compositor:
        _run_command(*stop_compositor, run_in_background=background)


def enable_compositing():
    """Re-enable compositing if the corresponding call to disable_compositing
    disabled it."""

    compositing_disabled = _COMPOSITING_DISABLED_STACK.pop()
    if not compositing_disabled:
        return
    start_compositor, _, background = _get_compositor_commands()
    if start_compositor:
        _run_command(*start_compositor, run_in_background=background)


class DBusScreenSaverInhibitor:
    """Inhibit and uninhibit the suspend using DBus.

    It will use the Gtk.Application's inhibit and uninhibit methods to
    prevent the computer from going to sleep.

    For enviroments which don't support either org.freedesktop.ScreenSaver or
    org.gnome.ScreenSaver interfaces one can declare a DBus interface which
    requires the Inhibit() and UnInhibit() methods to be exposed."""

    def __init__(self):
        self.proxy = None
        self._used_gtk_fallback = False

    def set_dbus_iface(self, name, path, interface, bus_type=Gio.BusType.SESSION):
        """Sets a dbus proxy to be used instead of Gtk.Application methods, this
        method can raise an exception."""
        self.proxy = Gio.DBusProxy.new_for_bus_sync(
            bus_type, Gio.DBusProxyFlags.NONE, None, name, path, interface, None
        )

    def inhibit(self, game_name):
        """Inhibit suspend.
        Returns a cookie that must be passed to the corresponding uninhibit() call.
        If an error occurs, None is returned instead."""
        reason = "Running game: %s" % game_name
        self._used_gtk_fallback = False

        if self.proxy:
            try:
                cookie = self.proxy.Inhibit("(ss)", "Lutris", reason)
                if cookie:
                    return cookie
                # D-Bus call succeeded but returned no cookie
                logger.warning("D-Bus screensaver inhibit returned no cookie, falling back to GTK")
            except Exception as ex:
                logger.warning("Failed to inhibit screensaver via D-Bus, falling back to GTK: %s", ex)

        # GTK fallback
        app = Gio.Application.get_default()
        window = app.window
        flags = Gtk.ApplicationInhibitFlags.SUSPEND | Gtk.ApplicationInhibitFlags.IDLE
        cookie = app.inhibit(window, flags, reason)

        # Gtk.Application.inhibit returns 0 if there was an error.
        if cookie == 0:
            return None

        self._used_gtk_fallback = True
        return cookie

    def uninhibit(self, cookie):
        """Uninhibit suspend.
        Takes a cookie as returned by inhibit. If cookie is None, no action is taken."""
        if not cookie:
            return

        if self.proxy and not self._used_gtk_fallback:
            try:
                self.proxy.UnInhibit("(u)", cookie)
                return
            except Exception as ex:
                logger.warning("Failed to uninhibit screensaver via D-Bus: %s", ex)

        # GTK fallback
        app = Gio.Application.get_default()
        app.uninhibit(cookie)


def _get_suspend_inhibitor():
    """Return the appropriate suspend inhibitor instance.
    If the required interface isn't available, it will default to GTK's
    implementation."""
    desktop_environment = get_desktop_environment()
    inhibitor = DBusScreenSaverInhibitor()

    # Build list of D-Bus interfaces to try in order of preference
    interfaces_to_try = []

    if desktop_environment is DesktopEnvironment.MATE:
        interfaces_to_try.append(("org.mate.ScreenSaver", "/", "org.mate.ScreenSaver"))
    elif desktop_environment is DesktopEnvironment.XFCE:
        # Try XFCE-specific interface first, then fall back to freedesktop.
        # Some XFCE setups use xfce4-screensaver (org.xfce.ScreenSaver),
        # others use light-locker or other lockers (org.freedesktop.ScreenSaver).
        interfaces_to_try.append(("org.xfce.ScreenSaver", "/", "org.xfce.ScreenSaver"))
        interfaces_to_try.append(
            ("org.freedesktop.ScreenSaver", "/org/freedesktop/ScreenSaver", "org.freedesktop.ScreenSaver")
        )

    for name, path, interface in interfaces_to_try:
        try:
            inhibitor.set_dbus_iface(name, path, interface)
            logger.debug("Using D-Bus screensaver interface: %s", name)
            return inhibitor
        except GLib.Error as err:
            logger.debug("D-Bus interface %s not available: %s", name, err)

    # No D-Bus interface available, will use GTK fallback
    return inhibitor


SCREEN_SAVER_INHIBITOR = _get_suspend_inhibitor()



"""Utilities for manipulating Wine"""

import os
from collections import OrderedDict
from gettext import gettext as _
from typing import Any, Dict, List, Optional, Set, Tuple

from lutris.exceptions import MisconfigurationError, UnspecifiedVersionError
from lutris.settings import WINE_DIR
from lutris.util import cache_single, linux, system
from lutris.util.log import logger
from lutris.util.strings import get_natural_sort_key, parse_version
from lutris.util.wine import fsync, proton

WINE_DEFAULT_ARCH: str = "win64" if linux.LINUX_SYSTEM.is_64_bit else "win32"
GE_PROTON_LATEST: str = "ge-proton"
WINE_PATHS: Dict[str, str] = {
    "winehq-devel": "/opt/wine-devel/bin/wine",
    "winehq-staging": "/opt/wine-staging/bin/wine",
    "wine-development": "/usr/lib/wine-development/wine",
    "system": "wine",
}

# Insert additional system-wide Wine installations.
try:
    if system.path_exists("/usr/lib"):
        for _candidate in os.listdir("/usr/lib/"):
            if _candidate.startswith("wine-"):
                _wine_path = os.path.join("/usr/lib/", _candidate, "bin/wine")
                if os.path.isfile(_wine_path):
                    WINE_PATHS["System " + _candidate] = _wine_path
    _candidate = None
    _wine_path = None
except Exception as ex:
    logger.exception("Unable to enumerate system Wine versions: %s", ex)


def detect_arch(prefix_path: Optional[str] = None, wine_path: Optional[str] = None) -> str:
    """Given a Wine prefix path, return its architecture"""
    if wine_path:
        if proton.is_proton_path(wine_path) or system.path_exists(wine_path + "64"):
            return "win64"
    if prefix_path and is_prefix_directory(prefix_path):
        return detect_prefix_arch(prefix_path)
    return "win32"


def is_prefix_directory(prefix_path: str) -> bool:
    """Detects if a path is ther oot of a Wine prefix; to be one, it must contain
    a 'system.reg' file."""
    if not prefix_path:
        return False

    prefix_path = os.path.expanduser(prefix_path)
    registry_path = os.path.join(prefix_path, "system.reg")
    return os.path.isdir(prefix_path) and os.path.isfile(registry_path)


def detect_prefix_arch(prefix_path: str) -> str:
    """Return the architecture of the prefix found in `prefix_path`"""
    if not is_prefix_directory(prefix_path):
        raise RuntimeError("Prefix not found: %s" % prefix_path)

    prefix_path = os.path.expanduser(prefix_path)
    registry_path = os.path.join(prefix_path, "system.reg")
    with open(registry_path, "r", encoding="utf-8") as registry:
        for _line_no in range(5):
            line = registry.readline()
            if "win64" in line:
                return "win64"
            if "win32" in line:
                return "win32"
    logger.error("Failed to detect Wine prefix architecture in %s; defaulting to 32-bit.", prefix_path)
    return "win32"


def set_drive_path(prefix: str, letter: str, path: str) -> None:
    """Changes the path to a Wine drive"""
    dosdevices_path = os.path.join(prefix, "dosdevices")
    if not system.path_exists(dosdevices_path):
        raise OSError("Invalid prefix path %s" % prefix)
    drive_path = os.path.join(dosdevices_path, letter + ":")
    if system.path_exists(drive_path):
        os.remove(drive_path)
    logger.debug("Linking %s to %s", drive_path, path)
    system.create_symlink(path, drive_path)


def is_gstreamer_build(wine_path: str) -> bool:
    """Returns whether a wine build ships with gstreamer libraries.
    This allows to set GST_PLUGIN_SYSTEM_PATH_1_0 for the builds that support it.
    """
    base_path = os.path.dirname(os.path.dirname(wine_path))
    return system.path_exists(os.path.join(base_path, "lib64/gstreamer-1.0"))


def is_winewayland_available(runner_version: str) -> bool:
    """Check if winewayland graphics driver available"""
    files_dir = get_runner_files_dir_for_version(runner_version)
    if files_dir:
        winewayland_paths = (
            "lib64/wine/x86_64-windows/winewayland.drv",
            "lib64/wine/i386-windows/winewayland.drv",
            "lib/wine/x86_64-windows/winewayland.drv",
            "lib/wine/i386-windows/winewayland.drv",
        )
        if not any(os.path.exists(os.path.join(files_dir, winewayland_path)) for winewayland_path in winewayland_paths):
            return False

    return True


def is_installed_systemwide() -> bool:
    """Return whether Wine is installed outside of Lutris"""
    for build in WINE_PATHS.values():
        if system.can_find_executable(build):
            return True
    return False


def list_system_wine_versions() -> List[str]:
    """Return the list of wine versions installed on the system"""
    versions = [name for name, path in WINE_PATHS.items() if get_system_wine_version(path)]
    return sorted(versions, key=get_natural_sort_key, reverse=True)


def list_lutris_wine_versions() -> List[str]:
    """Return the list of wine versions installed by lutris"""
    if not system.path_exists(WINE_DIR):
        return []
    versions = []
    for dirname in version_sort(os.listdir(WINE_DIR), reverse=True):
        try:
            wine_path = os.path.join(WINE_DIR, dirname, "bin/wine")
            if wine_path and os.path.isfile(wine_path):
                versions.append(dirname)
        except MisconfigurationError:
            pass  # if it's not properly installed, skip it
    return sorted(versions, key=get_natural_sort_key, reverse=True)


@cache_single
def get_installed_wine_versions() -> List[str]:
    """Return the list of Wine versions installed, with no duplicates and in
    the presentation order."""
    versions: Set[str] = {
        GE_PROTON_LATEST,
    }

    for v in proton.list_proton_versions():
        if v not in versions:
            versions.add(v)

    for v in list_lutris_wine_versions():
        if v not in versions:
            versions.add(v)

    for v in list_system_wine_versions():
        if v not in versions:
            versions.add(v)

    return list(versions)


def clear_wine_version_cache() -> None:
    get_installed_wine_versions.cache_clear()
    proton.get_proton_versions.cache_clear()
    proton.get_umu_path.cache_clear()


def get_runner_files_dir_for_version(version: str) -> Optional[str]:
    """This returns the path to the root of the Wine files for a specific version. The
    'bin' directory for that version is there, and we can place more directories there.
    If we shouldn't do that, this will return None."""
    if version == GE_PROTON_LATEST:
        return None
    if version in WINE_PATHS:
        return None
    elif proton.is_proton_version(version):
        return os.path.join(WINE_DIR, version, "files")
    else:
        return os.path.join(WINE_DIR, version)


def get_wine_path_for_version(version: str, config: Optional[dict] = None) -> str:
    """Return the absolute path of a wine executable for a given version,
    or the configured version if you don't ask for a version."""
    if not version and config:
        version = config["version"]

    if not version:
        raise UnspecifiedVersionError(_("The Wine version must be specified."))

    if version == GE_PROTON_LATEST:
        return proton.get_umu_path()
    if version in WINE_PATHS:
        return system.find_required_executable(WINE_PATHS[version])
    if proton.is_proton_version(version):
        return proton.get_proton_wine_path(version)
    if version == "custom":
        if config is None:
            raise RuntimeError("Custom wine paths are only supported when a configuration is available.")
        wine_path = config.get("custom_wine_path")
        if not wine_path:
            raise RuntimeError("The 'custom' Wine version can be used only if the custom wine path is set.")
        return wine_path
    return os.path.join(WINE_DIR, version, "bin/wine")


def parse_wine_version(version: str) -> Tuple[List[int], str, str]:
    """This is a specialized parse_version() that adjusts some odd
    Wine versions for correct parsing."""
    version = version.replace("Proton7-", "Proton-7.")
    version = version.replace("Proton8-", "Proton-8.")
    version = version.replace("Proton9-", "Proton-9.")
    return parse_version(version)


def version_sort(versions: List[str], reverse: bool = False) -> List[str]:
    def version_key(version: str) -> List[Any]:
        version_list, prefix, suffix = parse_wine_version(version)
        # Normalize the length of sub-versions
        sort_key: List[Any] = list(version_list) + [0] * (10 - len(version_list))
        sort_key.append(prefix)
        sort_key.append(suffix)
        return sort_key

    return sorted(versions, key=version_key, reverse=reverse)


def is_esync_limit_set() -> bool:
    """Checks if the number of files open is acceptable for esync usage."""
    return linux.LINUX_SYSTEM.has_enough_file_descriptors()


def is_fsync_supported() -> bool:
    """Checks if the running kernel has Valve's futex patch applied."""
    return fsync.get_fsync_support()


def get_default_wine_version() -> str:
    """Return the default version of wine."""
    return GE_PROTON_LATEST


def get_system_wine_version(wine_path: str = "wine") -> str:
    """Return the version of Wine installed on the system."""
    if wine_path != "wine" and not system.path_exists(wine_path):
        return ""
    if wine_path == "wine" and not system.can_find_executable("wine"):
        return ""
    version = system.read_process_output([wine_path, "--version"])
    if not version:
        logger.error("Error reading wine version for %s", wine_path)
        return ""
    if version.startswith("wine-"):
        version = version[5:]
    return version


def get_real_executable(windows_executable: str, working_dir: Optional[str]) -> Tuple[str, List[str], Optional[str]]:
    """Given a Windows executable, return the real program
    capable of launching it along with necessary arguments."""

    exec_name = windows_executable.lower()

    if exec_name.endswith(".msi"):
        return ("msiexec", ["/i", windows_executable], working_dir)

    if exec_name.endswith(".bat"):
        if not working_dir or os.path.dirname(windows_executable) == working_dir:
            working_dir = os.path.dirname(windows_executable)
            windows_executable = os.path.basename(windows_executable)
        return ("cmd", ["/C", windows_executable], working_dir)

    if exec_name.endswith(".lnk"):
        return ("start", ["/unix", windows_executable], working_dir)

    return (windows_executable, [], working_dir)


def get_overrides_env(overrides: Dict[str, str]) -> str:
    """
    Output a string of dll overrides usable with WINEDLLOVERRIDES
    See: https://wiki.winehq.org/Wine_User%27s_Guide#WINEDLLOVERRIDES.3DDLL_Overrides
    """
    default_overrides = {"winemenubuilder": ""}
    overrides.update(default_overrides)
    override_buckets: Dict[str, List] = OrderedDict(
        [("n,b", []), ("b,n", []), ("b", []), ("n", []), ("d", []), ("", [])]
    )
    for dll, value in overrides.items():
        if not value:
            value = ""
        value = value.replace(" ", "")
        value = value.replace("builtin", "b")
        value = value.replace("native", "n")
        value = value.replace("disabled", "")
        try:
            override_buckets[value].append(dll)
        except KeyError:
            logger.error("Invalid override value %s", value)
            continue

    override_strings = []
    for value, dlls in override_buckets.items():
        if not dlls:
            continue
        override_strings.append("{}={}".format(",".join(sorted(dlls)), value))
    return ";".join(override_strings)



"""Wine prefix management"""

import os

from lutris.settings import get_lutris_directory_settings, set_lutris_directory_settings
from lutris.util import joypad, system
from lutris.util.display import DISPLAY_MANAGER
from lutris.util.log import logger
from lutris.util.wine.registry import WineRegistry
from lutris.util.xdgshortcuts import get_xdg_entry

DESKTOP_KEYS = [
    "Desktop",
    "Personal",
    "My Music",
    "My Videos",
    "My Pictures",
    "{374DE290-123F-4565-9164-39C4925E467B}",  # Downloads
    "Templates",
]
DEFAULT_DESKTOP_FOLDERS = ["Desktop", "Documents", "Music", "Videos", "Pictures", "Downloads"]
DESKTOP_XDG = ["DESKTOP", "DOCUMENTS", "MUSIC", "VIDEOS", "PICTURES", "DOWNLOADS"]
DEFAULT_DLL_OVERRIDES = {
    "winemenubuilder": "",
}


def is_prefix(path):
    """Return True if the path is prefix"""
    return os.path.isdir(os.path.join(path, "drive_c")) and os.path.exists(os.path.join(path, "user.reg"))


def find_prefix(path):
    """Given an executable path, try to find a Wine prefix associated with it."""
    dir_path = path
    if not dir_path:
        logger.info("No path given, unable to guess prefix location")
        return
    dir_path = os.path.expanduser(dir_path)
    while dir_path != "/" and dir_path:
        dir_path = os.path.dirname(dir_path)
        if is_prefix(dir_path):
            return dir_path
        for prefix_dir in ("prefix", "pfx"):
            prefix_path = os.path.join(dir_path, prefix_dir)
            if is_prefix(prefix_path):
                return prefix_path


class WinePrefixManager:
    """Class to allow modification of Wine prefixes without the use of Wine"""

    hkcu_prefix = "HKEY_CURRENT_USER"
    hklm_prefix = "HKEY_LOCAL_MACHINE"

    def __init__(self, path):
        if not path:
            logger.warning("No path specified for Wine prefix")
        # expanduser() just in case- it should already be expanded.
        self.path = os.path.expanduser(path)

    def get_user_dir(self, default_user=None):
        user = default_user or os.getenv("USER") or "lutrisuser"
        return os.path.join(self.path, "drive_c/users/", user)

    @property
    def user_dir(self):
        """Returns the directory that contains the current user's profile in the WINE prefix."""
        return self.get_user_dir()

    @property
    def appdata_dir(self):
        """Returns the app-data directory for the user; this depends on a registry key."""
        user_dir = self.get_user_dir()
        folder = self.get_registry_key(
            self.hkcu_prefix + "/Software/Microsoft/Windows/CurrentVersion/Explorer/Shell Folders",
            "AppData",
        )
        if folder is None:
            logger.warning("Get Registry Key function returned NoneType to variable folder.")
        else:
            # Don't try to resolve the Windows path we get- there's
            # just two options, the Vista-and later option and the
            # XP-and-earlier option.
            if folder.lower().endswith("\\application data"):
                return os.path.join(user_dir, "Application Data")  # Windows XP
        return os.path.join(user_dir, "AppData/Roaming")  # Vista

    def setup_defaults(self):
        """Sets the defaults for newly created prefixes"""
        for dll, value in DEFAULT_DLL_OVERRIDES.items():
            self.override_dll(dll, value)

    def create_user_symlinks(self):
        """Link together user profiles created by Wine and Proton"""
        wine_user_dir = self.get_user_dir()
        proton_user_dir = self.get_user_dir(default_user="steamuser")
        if system.path_exists(wine_user_dir) and not system.path_exists(proton_user_dir, check_symlinks=True):
            system.create_symlink(wine_user_dir, proton_user_dir)
        elif system.path_exists(proton_user_dir) and not system.path_exists(wine_user_dir, check_symlinks=True):
            system.create_symlink(proton_user_dir, wine_user_dir)

    def get_registry_path(self, key):
        """Matches registry keys to a registry file

        Currently, only HKEY_CURRENT_USER keys are supported.
        """
        if key.startswith(self.hkcu_prefix):
            return os.path.join(self.path, "user.reg")
        if key.startswith(self.hklm_prefix):
            return os.path.join(self.path, "system.reg")
        raise ValueError("Unsupported key '{}'".format(key))

    def get_key_path(self, key):
        for prefix in (self.hkcu_prefix, self.hklm_prefix):
            if key.startswith(prefix):
                return key[len(prefix) + 1 :]
        raise ValueError("The key {} is currently not supported by WinePrefixManager".format(key))

    def get_registry_key(self, key, subkey):
        registry = WineRegistry(self.get_registry_path(key))
        return registry.query(self.get_key_path(key), subkey)

    def set_registry_key(self, key, subkey, value):
        registry = WineRegistry(self.get_registry_path(key))
        registry.set_value(self.get_key_path(key), subkey, value)
        registry.save()

    def clear_registry_key(self, key):
        registry = WineRegistry(self.get_registry_path(key))
        registry.clear_key(self.get_key_path(key))
        registry.save()

    def clear_registry_subkeys(self, key, subkeys):
        registry = WineRegistry(self.get_registry_path(key))
        registry.clear_subkeys(self.get_key_path(key), subkeys)
        registry.save()

    def override_dll(self, dll, mode):
        key = self.hkcu_prefix + "/Software/Wine/DllOverrides"
        if mode.startswith("dis"):
            mode = ""
        if mode not in ("builtin", "native", "builtin,native", "native,builtin", ""):
            logger.error("DLL override '%s' mode is not valid", mode)
            return
        self.set_registry_key(key, dll, mode)

    def get_desktop_folders(self):
        """Return the list of desktop folder names loaded from the Windows registry"""
        desktop_folders = []
        for key in DESKTOP_KEYS:
            folder = self.get_registry_key(
                self.hkcu_prefix + "/Software/Microsoft/Windows/CurrentVersion/Explorer/Shell Folders",
                key,
            )
            if not folder:
                logger.warning("Couldn't load shell folder name for %s", key)
                continue
            desktop_folders.append(folder[folder.rfind("\\") + 1 :])
        return desktop_folders or DEFAULT_DESKTOP_FOLDERS

    def install_desktop_integration(self):
        """Replace WINE's desktop folders with links to the corresponding
        folders in your home directory."""
        user_dir = self.user_dir
        home_dir = os.path.expanduser("~")
        current_dir = self._get_desktop_integration_assignment()

        if system.path_exists(user_dir, check_symlinks=True) and current_dir != home_dir:
            desktop_folders = self.get_desktop_folders()
            for i, item in enumerate(desktop_folders):
                path = os.path.join(user_dir, item)
                safe_path = path + ".winecfg"

                self._remove_desktop_folder(path, safe_path)

                # if we want to create a symlink and one is already there, just
                # skip to the next item.  this also makes sure we don't
                # find a dir (isdir only looks at the target of the symlink).
                src_path = get_xdg_entry(DESKTOP_XDG[i])
                if not src_path:
                    logger.error("No XDG entry found for %s, launcher not created", DESKTOP_XDG[i])
                else:
                    system.create_symlink(src_path, path)

            self._set_desktop_integration_assignment(home_dir)

    def remove_desktop_integration(self):
        """Replace the desktop integration links with proper folders."""
        user_dir = self.user_dir
        current_dir = self._get_desktop_integration_assignment()

        if system.path_exists(user_dir) and current_dir != user_dir:
            desktop_folders = self.get_desktop_folders()
            for item in desktop_folders:
                path = os.path.join(user_dir, item)
                safe_path = path + ".winecfg"

                # Disintegration means the desktop folders in WINE are
                # actual directories not links, and that's all we want.
                if not os.path.islink(path):
                    continue

                self._remove_desktop_folder(path, safe_path)

                # We prefer to restore the previously saved directory.
                if os.path.isdir(safe_path):
                    os.rename(safe_path, path)
                else:
                    os.makedirs(path, exist_ok=True)

            self._set_desktop_integration_assignment(user_dir)

    def _remove_desktop_folder(self, path, safe_path):
        """Removes the link or directory at 'path'; if it is a non-empty directory
        this will rename it to 'safe_path' instead of removing it entirely."""
        if os.path.islink(path):
            os.unlink(path)
        elif os.path.isdir(path):
            try:
                os.rmdir(path)
            except OSError:
                # We can't delete nonempty dir, so we rename as wine do.
                os.rename(path, safe_path)

    def _get_desktop_integration_assignment(self):
        try:
            # If the old tracking file is found, we'll read it, unlink it, and
            # save the setting in the new form.
            obsolete_path = os.path.join(self.path, ".lutris_destkop_integration")
            if os.path.isfile(obsolete_path):
                with open(obsolete_path, "r", encoding="utf-8") as f:
                    desktop_dir = f.read()
                self._set_desktop_integration_assignment(desktop_dir)
                os.unlink(obsolete_path)
        except Exception as ex:
            logger.exception("Unable to read Lutris desktop integration setting: %s", ex)

        settings = get_lutris_directory_settings(self.path)
        return settings.get("desktop_integration_directory", "")

    def _set_desktop_integration_assignment(self, desktop_dir):
        set_lutris_directory_settings(self.path, {"desktop_integration_directory": desktop_dir or ""})

    def set_crash_dialogs(self, enabled):
        """Enable or diable Wine crash dialogs"""
        self.set_registry_key(
            self.hkcu_prefix + "/Software/Wine/WineDbg",
            "ShowCrashDialog",
            1 if enabled else 0,
        )

    def set_virtual_desktop(self, enabled):
        """Enable or disable wine virtual desktop.
        The Lutris virtual desktop is refered to as 'WineDesktop', in Wine the
        virtual desktop name is 'default'.
        """
        path = self.hkcu_prefix + "/Software/Wine/Explorer"
        if enabled:
            self.set_registry_key(path, "Desktop", "WineDesktop")
            default_resolution = "x".join(DISPLAY_MANAGER.get_current_resolution())
            logger.debug(
                "Enabling wine virtual desktop with default resolution of %s",
                default_resolution,
            )
            self.set_registry_key(
                self.hkcu_prefix + "/Software/Wine/Explorer/Desktops",
                "WineDesktop",
                default_resolution,
            )
        else:
            self.clear_registry_key(path)

    def set_desktop_size(self, desktop_size):
        """Sets the desktop size if one is given but do not reset the key if
        one isn't.
        """
        path = self.hkcu_prefix + "/Software/Wine/Explorer/Desktops"
        if desktop_size:
            self.set_registry_key(path, "WineDesktop", desktop_size)

    def set_dpi(self, dpi):
        """Sets the DPI for WINE to use. None to remove the Lutris setting,
        and leave WINE in control."""

        # Convert the old hidden file into a 'lutris.json' settings file
        obsolete_path = os.path.join(self.path, ".lutris_dpi_assignment")
        try:
            if os.path.isfile(obsolete_path):
                with open(obsolete_path, "r", encoding="utf-8") as f:
                    dpi_assigned = int(f.read())
                set_lutris_directory_settings(self.path, int(dpi_assigned))
                os.unlink(obsolete_path)
        except Exception as ex:
            logger.exception("Unable to read Lutris assigned DPI: %s", ex)

        settings = get_lutris_directory_settings(self.path)

        key_paths = [self.hkcu_prefix + "/Software/Wine/Fonts", self.hkcu_prefix + "/Control Panel/Desktop"]

        def assign_dpi(dpi):
            for key_path in key_paths:
                self.set_registry_key(key_path, "LogPixels", dpi)

        def is_lutris_dpi_assigned():
            """Check if Lutris assigned the DPI presently found in the registry."""
            try:
                dpi_assigned = settings.get("dpi_assigned")
                if dpi_assigned:
                    dpi_assigned = int(dpi_assigned)
                else:
                    return False
            except Exception as ex:
                logger.exception("Unable to read Lutris assigned DPI: %s", ex)
                return False

            for key_path in key_paths:
                if dpi_assigned != self.get_registry_key(key_path, "LogPixels"):
                    return False
            return True

        if dpi:
            assign_dpi(dpi)
            set_lutris_directory_settings(self.path, {"dpi_assigned": dpi})
        elif settings.get("dpi_assigned"):
            if is_lutris_dpi_assigned():
                assign_dpi(96)  # reset previous DPI
            set_lutris_directory_settings(self.path, {"dpi_assigned": ""})

    def cleanup_broken_symlinks(self):
        """Remove broken symlinks from system32/syswow64 before Proton runs.

        When switching Proton versions, DLL symlinks from the previous version
        remain in the prefix pointing to a now-uninstalled Proton. Proton's
        pfx_copy calls os.symlink() without checking if the destination exists,
        causing FileExistsError on prefix upgrades.
        """
        removed_count = 0

        # prefix_path may be the compatdata directory with pfx/ inside it
        pfx_path = os.path.join(self.path, "pfx")
        prefix = pfx_path if os.path.isdir(pfx_path) else self.path

        for subdir in ("system32", "syswow64"):
            dir_path = os.path.join(prefix, "drive_c", "windows", subdir)
            if not os.path.isdir(dir_path):
                continue
            try:
                entries = os.listdir(dir_path)
            except OSError as ex:
                logger.warning("Failed to list %s: %s", dir_path, ex)
                continue
            for entry in entries:
                full_path = os.path.join(dir_path, entry)
                if os.path.islink(full_path) and not os.path.exists(full_path):
                    try:
                        os.unlink(full_path)
                        removed_count += 1
                    except OSError as ex:
                        logger.warning("Failed to remove broken symlink %s: %s", full_path, ex)

        if removed_count:
            logger.debug("Removed %d broken symlink(s) from prefix", removed_count)

    def configure_joypads(self):
        """Disables some joypad devices"""
        key = self.hkcu_prefix + "/Software/Wine/DirectInput/Joysticks"
        self.clear_registry_key(key)
        for _device, joypad_name in joypad.get_joypads():
            # Attempt at disabling mice that register as joysticks.
            # Although, those devices aren't returned by `get_joypads`
            # A better way would be to read /dev/input files directly.
            if "HARPOON RGB" in joypad_name:
                self.set_registry_key(key, "{} (js)".format(joypad_name), "disabled")
                self.set_registry_key(key, "{} (event)".format(joypad_name), "disabled")

        # This part of the code below avoids having 2 joystick interfaces
        # showing up simulatenously. It is not sure if it's still needed
        # so it is disabled for now. Street Fighter IV now runs in Proton
        # without this sort of hack.
        #
        # for device, joypad_name in joypads:
        #     if "event" in device:
        #         disabled_joypad = "{} (js)".format(joypad_name)
        #     else:
        #         disabled_joypad = "{} (event)".format(joypad_name)
        #     self.set_registry_key(key, disabled_joypad, "disabled")



# Standard Library
from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError

# Lutris Modules
from lutris.runners.runner import Runner
from lutris.util import system


class pcsx2(Runner):
    human_name = _("PCSX2")
    description = _("PlayStation 2 emulator")
    platforms = [_("Sony PlayStation 2")]
    runnable_alone = True
    runner_executable = "pcsx2/PCSX2"
    flatpak_id = "net.pcsx2.PCSX2"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ISO file"),
            "default_path": "game_path",
        }
    ]

    runner_options = [
        {
            "option": "fullscreen",
            "type": "bool",
            "label": _("Fullscreen"),
            "default": False,
        },
        {"option": "full_boot", "type": "bool", "label": _("Fullboot"), "default": False},
        {"option": "nogui", "type": "bool", "label": _("No GUI"), "default": False},
    ]

    # PCSX2 currently uses an AppImage, no need for the runtime.
    system_options_override = [{"option": "disable_runtime", "default": True}]

    def play(self):
        arguments = self.get_command()

        if self.runner_config.get("fullscreen"):
            arguments.append("-fullscreen")
        if self.runner_config.get("full_boot"):
            arguments.append("-slowboot")
        if self.runner_config.get("nogui"):
            arguments.append("-nogui")

        iso = self.game_config.get("main_file") or ""
        if not system.path_exists(iso):
            raise MissingGameExecutableError(filename=iso)
        arguments.append(iso)
        return {"command": arguments}



"""Steam appmanifest file handling"""

import os
import re

from lutris.util.log import logger
from lutris.util.steam.config import get_steamapps_dirs
from lutris.util.steam.vdfutils import vdf_parse
from lutris.util.strings import slugify
from lutris.util.system import fix_path_case, path_exists

APP_STATE_FLAGS = [
    "Invalid",
    "Uninstalled",
    "Update Required",
    "Fully Installed",
    "Encrypted",
    "Locked",
    "Files Missing",
    "AppRunning",
    "Files Corrupt",
    "Update Running",
    "Update Paused",
    "Update Started",
    "Uninstalling",
    "Backup Running",
    "Reconfiguring",
    "Validating",
    "Adding Files",
    "Preallocating",
    "Downloading",
    "Staging",
    "Committing",
    "Update Stopping",
]


class AppManifest:
    """Representation of an AppManifest file from Steam"""

    def __init__(self, appmanifest_path):
        self.appmanifest_path = appmanifest_path
        self.steamapps_path, filename = os.path.split(appmanifest_path)
        self.steamid = re.findall(r"(\d+)", filename)[-1]
        self.appmanifest_data = {}

        if path_exists(appmanifest_path):
            with open(appmanifest_path, "r", encoding="utf-8") as appmanifest_file:
                self.appmanifest_data = vdf_parse(appmanifest_file, {})
        else:
            logger.error("Path to AppManifest file %s doesn't exist", appmanifest_path)

    def __repr__(self):
        return "<AppManifest: %s>" % self.appmanifest_path

    @property
    def app_state(self):
        """State of the app (dictionary containing game specific info)"""
        return self.appmanifest_data.get("AppState") or {}

    @property
    def user_config(self):
        """Return the user configuration part"""
        return self.app_state.get("UserConfig") or {}

    @property
    def name(self):
        """Return the game name from either the state or the user config"""
        _name = self.app_state.get("name")
        if not _name:
            _name = self.user_config.get("name")
        return _name

    @property
    def slug(self):
        """Return a slugified version of the name"""
        return slugify(self.name)

    @property
    def installdir(self):
        """Path where the game is installed"""
        return self.app_state.get("installdir")

    @property
    def states(self):
        """Return the states of a Steam game."""
        states = []
        state_flags = self.app_state.get("StateFlags", 0)
        state_flags = bin(int(state_flags))[:1:-1]
        for index, flag in enumerate(state_flags):
            if flag == "1":
                states.append(APP_STATE_FLAGS[index + 1])
        return states

    def is_installed(self):
        """True if the game is fully installed"""
        return "Fully Installed" in self.states

    def get_install_path(self):
        """Absolute path of the installation directory"""
        if not self.installdir:
            return None
        install_path = fix_path_case(os.path.join(self.steamapps_path, "common", self.installdir))
        if install_path and path_exists(install_path):
            return install_path
        return None

    def get_platform(self):
        """Platform the game uses (linux or windows)"""
        steamapps_paths = get_steamapps_dirs()
        if self.steamapps_path in steamapps_paths["linux"]:
            return "linux"
        if self.steamapps_path in steamapps_paths["windows"]:
            return "windows"
        raise ValueError("Can't find %s in %s" % (self.steamapps_path, steamapps_paths))


def get_appmanifest_from_appid(steamapps_path, appid):
    """Given the steam apps path and appid, return the corresponding appmanifest"""
    if not steamapps_path:
        raise ValueError("steamapps_path is mandatory")
    if not path_exists(steamapps_path):
        raise IOError("steamapps_path must be a valid directory")
    if not appid:
        raise ValueError("Missing mandatory appid")
    appmanifest_path = os.path.join(steamapps_path, "appmanifest_%s.acf" % appid)
    if not path_exists(appmanifest_path):
        return None
    return AppManifest(appmanifest_path)


def get_path_from_appmanifest(steamapps_path, appid):
    """Return the path where a Steam game is installed."""
    appmanifest = get_appmanifest_from_appid(steamapps_path, appid)
    if not appmanifest:
        return None
    return appmanifest.get_install_path()


def get_appmanifests(steamapps_path):
    """Return the list for all appmanifest files in a Steam library folder"""
    return [f for f in os.listdir(steamapps_path) if re.match(r"^appmanifest_\d+.acf$", f)]



import os

from lutris.game import Game


def generate_script(logger, launch_ui_delegate, db_game, script_path):
    """Output a script to a file.
    The script is capable of launching a game without the client
    """

    logger.info(f"Try creating script for '{db_game['name']}'")

    def on_error(error: BaseException) -> None:
        logger.exception("Unable to generate script: %s", error)

    game = Game(db_game["id"])
    game.game_error.register(on_error)
    game.reload_config()
    game.write_script(script_path, launch_ui_delegate)
    absolutePath = os.path.abspath(script_path)
    logger.info(f"Wrote script to: '{absolutePath}'")
